/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file contains the class for processing XML configuration files
** (read/write).
**
****************************************************************************/
#pragma once

#include "stdint.h"
#include "interface/studio_setup_structs.h"
#include "tinyxml2/tinyxml2.h"
#include "common/software_version.h"

#include <memory>


/** Klasse zur Erstellung und Verarbeitung von XML-Konfigurationsdateien.
*/
class FileXML
{
public:
    enum class ErrorCodes
    {
        Success = 0,
        LoadingHelpFileFailed,
        DocumentCreationFailed,
        FileCreationFailed,
        FileSaveFailed,
    };

    enum class ContentType
    {
        MainNode = 0,
        FileHeader,
        FlashHeader,
        ImageScope,
        ExtendedScope,
        Config,
        GlobalVar,
        GlobalVarValue,
        ConfigRoot,
        Operating,
        Menu,
        MenuItem,
        ReadRecord,
        ReadField,
        ReadField_TextEx,
        ReadField_ConfirmationEx,
        ReadField_NumFieldEx,
        ReadField_ListSelectionEx,
        ReadField_DataEx,
        ReadField_GlobVarToHIDEx,
        ReadField_SwitchModuleEnergieEx,
        ReadField_SystemMessageEx,
        ReadField_TakePhotoEx,
        ReadField_FeverControl,
        ReadField_MultiInputSource,
        Transponder,
        TransponderGlobal,
        TransponderAccessControl,
        Compare,
        Records,
        RecordDescription,
        RecordField,
        Lists,
        ExtendedFlash,
        MachineDefinitions,
        Machine,
        VibrationSensor,
        SignalProcessing,
        DigitalInputs,
        DigitalInput,
        DigitalInputEx,
        AnalogInputs,
        AnalogInput,
        Timers,
        Timer,
        EventChains,
        IOModule,
        AccessControl,
        Access,
        PowerSetupNode,
        PowerSetup,
        PowerSetup_CPU,
        PowerSetup_Timer,
        PowerSetup_Inputs,
        PowerSetup_DeviceGlobal,
        PowerSetup_U5V,
        PowerSetup_U12V,
        PowerSetup_Communication,
        PowerSetup_DisplayLight,
        PowerSetup_DisplayText,
        PowerSetup_Fingerprint,
        PowerSetup_Transponder,
        PowerSetup_GPS,
        PowerSetup_USBHost,
        PowerSetup_MMC,
        PowerEvent,
        PowerMemo,
        TransponderConfiguration,
        TransponderLegicAdvantHeader,
        TransponderLegicAdvant,
        TransponderMifareDesfireHeader,
        TransponderMifareDesfire,
        TransponderMifarePlusHeader,
        TransponderMifarePlus,
        TransponderMifareUltralightHeader,
        TransponderMifareUltralight,
        TransponderiCodeHeader,
        TransponderiCode,
        TransponderJavaCardHeader,
        TransponderJavaCard,
        TransponderMyDHeader,
        TransponderMyD,
        TransponderISOTagsHeader,
        TransponderISOTags,
        TransponderWiegandClockDataHeader,
        TransponderWiegandClockData,
        TransponderETSITagsHeader,
        TransponderETSITags,
        TransponderQ5Header,
        TransponderQ5,
        TransponderHitag1FeigHeader,
        TransponderHitag1Feig,
        TransponderPreconfiguredReaderHeader,
        TransponderPreconfiguredReader,
        Extended,
        Fingerprint,
        TimeboyAdaption,
        GPRS,
        EnergieOption,
        GPS,
        ProtectionPassword,
        AutoKeys,
        AutoKey,
        SystemMessages,
        SystemMessage
    };

public:
    FileXML();

    std::shared_ptr<tinyxml2::XMLDocument> xmlDoc() const
    {
        return m_xmlDoc;
    }

    tinyxml2::XMLElement *parentNode() const
    {
        return m_parent;
    }

    tinyxml2::XMLElement *setParentNode( tinyxml2::XMLElement *newParent )
    {
        auto oldParent = m_parent;
        m_parent = newParent;
        return oldParent;
    }

    void setVersions( const std::string &studioVersion, const std::string &libraryVersion );
    FileXML::ErrorCodes init();
    FileXML::ErrorCodes reset();
    FileXML::ErrorCodes save( const std::string &filePath );
    void writeHelpFile( const std::string &path );

    void setValidState( bool value )
    {
        m_valid = value;
    }

    void combineValid( bool value )
    {
        m_valid &= value;
    }

    bool isValid()
    {
        return m_valid;
    }

    SoftwarebuildVersion &studioVersion()
    {
        return m_studioVersion;
    }

    SoftwarebuildVersion &libraryVersion()
    {
        return m_libraryVersion;
    }

    void configureAdditionalHelpFile( bool createHelpFile, const std::string &path );

    /**Sucht den Elternknoten anhand eines definierten Inhaltstypes und gibt einen Zeiger darauf zurück.

    Sollte der Knoten nicht vorhanden sein, so wird dieser erstellt und zurückgegeben. Wenn mehrere Knoten mit
    dem selben Namen vorhanden sind, wird immer der Letzte verwendet.

    @return Zeiger auf den gewünschten Knoten
    */
    tinyxml2::XMLElement *nodeByType( ContentType type );

    const std::string &nodeNameByType( ContentType type ) const;
    unsigned int numberOfChildItems( ContentType type );

    template<typename T>
    void writeContent( ContentType type, const T &data );

    template<typename T>
    bool updateContent( ContentType type, const T &data, tinyxml2::XMLElement *&lastItem );

    template<typename T>
    bool readContent( ContentType type, T &data, tinyxml2::XMLElement *&lastItem );

    template<typename T>
    bool readContent( ContentType type, T &data, const uint32_t index, tinyxml2::XMLElement *&lastItem )
    {
        const std::string &nodeName = nodeNameByType( type );
        bool errorPresent = ((m_parent == nullptr) || nodeName.empty());
        assert( !errorPresent );
        if ( !errorPresent )
        {
            if ( index > 0 )
            {
                lastItem = m_parent->FirstChildElement( nodeName.c_str() );
                for ( uint32_t i = 1; i < index; ++i )
                {
                    lastItem = lastItem->NextSiblingElement( nodeName.c_str() );
                    // not enough items available
                    if ( lastItem == nullptr )
                    {
                        return false;
                    }
                }
            }
            return readContent( type, data, lastItem );
        }
        return false;
    }

    template<typename T>
    bool readContent( ContentType type, T &data )
    {
        tinyxml2::XMLElement *lastItem{};
        return readContent( type, data, lastItem );
    }

    void writeFileNames( const std::array<std::string, 20> &listFileNames, const std::array<std::string, 20> &listEntrance2FileNames );
    bool readFileNames( std::array<std::string, 20> &listFileNames, std::array<std::string, 20> &listEntrance2FileNames );

private:
    /** Konvertiert den XML-Tag Namen in den eigentlichen Strukturbezeichner.

    Die Beziehung zwischen Tag-Namen und Strukturbezeichner definiert sich folgendermaßen:
    Bezeichner = "Struct" + [Tag-Name] (erste Buchstabe des Tag-Names wird großgeschrieben)

    @param tag Name des XML-Tags
    @return Strukturbezeichner
    */
    std::string tagToStruct( const std::string &tag );
    const tinyxml2::XMLNode *findStructCompound( const std::string &name ) const;
    void processItem( std::ofstream &helpFile, tinyxml2::XMLElement *element, std::string shifting, const tinyxml2::XMLNode *lastValidCompound = nullptr );
    const char *changeSpecialTag( const char *name );

    void writeValue( tinyxml2::XMLElement *node, const std::string &name, const char *value );
    void writeValue( tinyxml2::XMLElement *node, const std::string &name, const uint8_t *value, int16_t size );

    void readValue( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem,
                    const std::string &name, uint8_t *value, uint16_t size,
                    bool &errorPresent, bool nullTermination = false );
    void readValue( tinyxml2::XMLElement *node,
                    const std::string &name, uint8_t *value, uint16_t size,
                    bool &errorPresent, bool nullTermination = false );

    void readValue( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem,
                    const std::string &name, char *value, uint16_t size,
                    bool &errorPresent, bool nullTermination = false );
    void readValue( tinyxml2::XMLElement *node,
                    const std::string &name, char *value, uint16_t size,
                    bool &errorPresent, bool nullTermination = false );

    template<typename T>
    void writeValue( tinyxml2::XMLElement *node, const std::string &name, const T &value );

    template<typename T>
    bool updateValue( tinyxml2::XMLElement *node, const std::string &name, const T &value )
    {
        if ( node != nullptr )
        {
            auto pItem = node->FirstChildElement( name.c_str() );
            if ( pItem != nullptr )
            {
                pItem->SetText( std::to_string( value ).c_str() );
                return true;
            }
        }
        return false;
    }

    template<typename T>
    T &readValue( tinyxml2::XMLElement *node, tinyxml2::XMLElement *&lastItem, const std::string &name, T &value, bool &errorPresent );

    template<typename T>
    T &readValue( tinyxml2::XMLElement *node, const std::string &name, T &value, bool &errorPresent )
    {
        tinyxml2::XMLElement *lastItem{};
        return readValue( node, lastItem, name, value, errorPresent );
    }

private:
    static std::string TAB;
    static std::string LT;
    static std::string GT;
    static std::string sTAG;
    static std::string eTAG;
    static std::map<ContentType, std::string> nodeNames;

    bool m_createHelpFile{};
    std::string m_docPath{};
    std::shared_ptr<tinyxml2::XMLDocument> m_pDoxygenTagXml{};  ///< doxygen tag file (needed for parsing the help information)
    tinyxml2::XMLElement *m_pDoxygenRoot{};                     ///< Wurzelknoten der Doxygen-Tag-Datei
    std::shared_ptr<tinyxml2::XMLDocument> m_xmlDoc{};          ///< XML-Dokument
    tinyxml2::XMLElement *m_parent{};                           ///< Elternknoten für die aktuelle Verarbeitung

    SoftwarebuildVersion m_minStudioVersion;
    SoftwarebuildVersion m_studioVersion;
    SoftwarebuildVersion m_libraryVersion;

    bool m_valid{ true };
};
