/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "channel/channel_bridge.h"
#include "DatafoxLibraryIV/include/dfcom_version.h"
#include "DatafoxLibraryIV/include/datafox_library_iv.h"
#include "common/eeprom.h"
#include "common/crc.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#if defined(_MSC_VER) || defined(__MINGW32__)
#include <io.h>
#else
#include <sys/uio.h>
#endif
#if !defined(DATAFOX_STUDIO_IV)
#include "file/file_studio_setup_xml.h"
#endif

#include "file/interface/studio_setup_classes.h"

#if defined(_MSC_VER) && !defined(snprintf)
#define snprintf _snprintf
#endif

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

ChannelBridge::ChannelBridge( int connectionNumber )
    : CDFLog( connectionNumber, std::bind( &ChannelBridge::changeTraceFilePath, this, std::placeholders::_1 ) )
{
    // Kanalnummer setzen
    setConnectionNumber( connectionNumber );

    // Protokollzustaende der angesprochenen Geraete
    memset( m_protocol_version, 0, sizeof( m_protocol_version ) );

    // Bevorzugt AEIII+ pruefen
    m_prefer_AEIII = false;

    // Protocolversion für späteren Pufferzugriff halten.
    m_protocol_version_record_description = 4;
    m_protocol_version_list_description = 4;

    // Initialisierungsdatei lesen
    logIniRead(); // führt auch logCheck durch.

    // Ist die bevorzugte Version der Dockingstation == 1, dann wird diese Klasse aktiviert.
    m_channel4.protocol().setTimeboyStationEnabled( m_dockingPreferType == 1 );

    // Datensatzpuffer loeschen
    m_recordCollected = false;
    memset( m_recordData, 0, sizeof( m_recordData ) );
    m_recordSize = 0;
    m_recordError = 0;

    // Wiederholungsversuche bei checkDevice
    m_device_poll_retry = 0;

    // Info zum binden des Kanals an ein bestimmtes Geraet.
    // memset( &m_bind_info, 0, sizeof(m_bind_info) );
    m_bindInfo.typeDevice = -1; // 0 wuerde PZE-MasterIV sein. -1 ist kein Geraet.

    m_channel4.setTraceMode( static_cast<TraceBase::TraceMode>( m_traceMode ) );
    m_channel4.initDeviceUpdate( m_dbPath, m_dbPathType );
}

ChannelBridge::~ChannelBridge()
{

}

bool ChannelBridge::tryEnterCriticalSection()
{
    return m_mutex_api.try_lock();
}

void ChannelBridge::enterCriticalSection()
{
    m_mutex_api.lock();
}

void ChannelBridge::leaveCriticalSection()
{
    m_mutex_api.unlock();
}

void ChannelBridge::changeTraceFilePath( const std::string &filePath )
{
    m_channel4.setTraceFilePath( filePath );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::openAEIII( int iType, const char *szComString, uint32_t iComValue, uint16_t iComTimeOut )
{
    // Bevorzugt AEIII+ pruefen
    m_prefer_AEIII = true;

    return open( ProtocolBase::defaultDeviceAddress(), iType, szComString, iComValue, iComTimeOut );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::open( uint8_t deviceNumber, int iType, const std::string &szComString, uint32_t iComValue, uint16_t iComTimeOut )
{
    int res = IV_ERROR_FAIL;

    // Kanal darf nicht gebunden sein.
    if ( isBind() )
    {
        return IV_ERROR_FAIL;
    }

    // Kanal 4 ist Massgebend
    if ( m_channel4.connection().type() == ConnectionBase::ConnectionType::undefined )
    {
        if ( m_channel4.connection().hostName() == szComString )
        {
            // Bestimmt die Gueltigkeit der Protokollversion

            // Alle Teilnehmer erhalten unbekanntes Protokoll
            memset( m_protocol_version, 0, sizeof( m_protocol_version ) );
            // Ggf. unbekannte Dockingstation
            m_channel4.protocol().resetTimeboyStation();
        }

        // Je nach Typ der Verbindung die Klasse initialisieren.
        switch ( iType )
        {
            case 1: // RS232 direkt (ohne autobaud)
                // Schnittstelle oeffnen
                m_channel4.setAutoBaud( false );
                res = m_channel4.openConnection( ConnectionBase::ConnectionType::serial, szComString, iComValue, iComTimeOut );
                break;

            case 3: // TCP/IP
                // Schnittstelle oeffnen
                m_channel4.setAutoBaud( false );
                res = m_channel4.openConnection( ConnectionBase::ConnectionType::socket, szComString, iComValue, iComTimeOut );
                break;

            case 4: // RS232 direkt mit autobaud
                // Schnittstelle oeffnen
                m_channel4.setAutoBaud( true );
                res = m_channel4.openConnection( ConnectionBase::ConnectionType::serial, szComString, iComValue, iComTimeOut );
                break;

            default: // Unbekannter Schnittstellentyp
                res = IV_ERROR_FAIL;
        }

        // Nach Erfolg die Grundeinstellung wiederherstellen
        if ( res == IV_ERROR_SUCCESS )
        {
            setDeviceNumberSender( deviceNumber );

            if ( m_channel4.connection().hostName() == szComString )
            {
                // Bestimmt die Gueltigkeit der Protokollversion

                // Alle Teilnehmer erhalten unbekanntes Protokoll
                memset( m_protocol_version, 0, sizeof( m_protocol_version ) );
                // Ggf. unbekannte Dockingstation
                m_channel4.protocol().resetTimeboyStation();
            }
        }
    }

    return res;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void ChannelBridge::close()
{
    // Kanal 4 ist Massgebend
    m_channel4.closeConnection();

    // Bevorzugt AEIII+ pruefen
    m_prefer_AEIII = false;
}

// ##################################################################
/**
* @brief Ermitteln ob die Schnittstelle geoeffnet ist.
*/
bool ChannelBridge::isOpen()
{
    // Channel4 ist Massgebend!
    return m_channel4.isConnectionOpen();
}

// ##################################################################
/**
* @brief Schnittstelle nach Fehler bereinigen.
*/
void ChannelBridge::resetWithError( int err )
{
    // Gehaltene Protokollversion auswerten
    switch ( m_protocol_version[(m_channel4.protocol().deviceNumberReceiver() == 0xFE) ? 0 : m_channel4.protocol().deviceNumberReceiver()] )
    {
        case 0:
        case 1:
        case 2:
        case 3: // Anstehende Daten verwerfen
            if ( err == IV_ERROR_CRC_PACKET || err == IV_ERROR_CHANNEL_COMMUNICATION || err == IV_ERROR_RECEIVE )
            {
                m_channel4.connection().flush();
            }
            BOOST_FALLTHROUGH;
        case 4:
            // Protocolversion zuruecksetzen
            if ( err == IV_ERROR_CHANNEL_COMMUNICATION || err == IV_ERROR_RECEIVE )
            {
                setProtocolVersion( deviceNumberReceiver(), 0 );
                // Ggf. unbekannte Dockingstation
                m_channel4.protocol().resetTimeboyStation();
            }
    }
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
uint8_t ChannelBridge::setDeviceNumberReceiver( uint8_t deviceNumber )
{
    uint8_t deviceNumberPrev = m_channel4.protocol().deviceNumberReceiver();

    m_channel1_2.SetDeviceIDRemote( deviceNumber );
    m_channel4.protocol().setDeviceNumberReceiver( deviceNumber );

    return deviceNumberPrev;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
uint8_t ChannelBridge::deviceNumberReceiver()
{
    return m_channel4.protocol().deviceNumberReceiver();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
uint8_t ChannelBridge::setDeviceNumberSender( uint8_t deviceNumber )
{
    uint8_t deviceAddressPrev = m_channel4.protocol().deviceNumberSender();

    m_channel1_2.SetDeviceID( deviceNumber );
    m_channel4.protocol().setDeviceNumberSender( deviceNumber );

    return deviceAddressPrev;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
uint8_t ChannelBridge::deviceNumberSender()
{
    return m_channel4.protocol().deviceNumberSender();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::comReleaseHandle()
{
    // CChannel4
    return m_channel4.connection().detach();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::comUseHandle( HANDLE hCom )
{
    // CChannel4
    return  m_channel4.connection().attach( ConnectionBase::ConnectionType::serial, hCom, 800 );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
HANDLE ChannelBridge::getComPort()
{
    HANDLE hCom = INVALID_HANDLE_VALUE;

    if ( m_channel4.connection().type() == ConnectionBase::ConnectionType::serial )
    {
        hCom = m_channel4.connection().serialPort();
    }
    else if ( m_channel4.connection().type() == ConnectionBase::ConnectionType::socket )
    {
        hCom = (HANDLE)m_channel4.connection().socket();
    }

    return hCom;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::setComPort( HANDLE hCom, int iType, uint32_t iComValue, uint16_t iTimeOut )
{
    switch ( iType )
    {
        case 0:
        case 2:
            // Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
            if ( (m_channel4.connection().type() == ConnectionBase::ConnectionType::undefined) ||
                ((m_channel4.connection().type() == ConnectionBase::ConnectionType::serial) && (m_channel4.connection().serialPort() == hCom)) )
            {
                if ( m_channel4.connection().changeSerialState( hCom, iComValue, iTimeOut ) != IV_ERROR_SUCCESS )
                {
                    return -IV_ERROR_FAIL;
                }

                // CChannel4
                m_channel4.connection().attach( ConnectionBase::ConnectionType::serial, hCom, iTimeOut );
            }
            else
            {
                return -IV_ERROR_FAIL;
            }
            break;

        case 1:
        case 3:
            // Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
            if ( (m_channel4.connection().type() == ConnectionBase::ConnectionType::undefined) ||
                ((m_channel4.connection().type() == ConnectionBase::ConnectionType::socket) && (m_channel4.connection().socket() == (SOCKET)hCom)) )
            {
                // CChannel4
                m_channel4.connection().attach( ConnectionBase::ConnectionType::socket, hCom, iTimeOut );
            }
            else
            {
                return -IV_ERROR_FAIL;
            }
            break;
    }

    return IV_ERROR_SUCCESS;
}

void ChannelBridge::setCallBackFunction( int (*pFunction)(void) )
{
    m_channel1_2.SetCBFunction( pFunction );
    m_channel4.protocol().setCallBackFunction( pFunction == nullptr ? std::function<int()>() : pFunction );
}

// Setzen des Logdateinamen.
void ChannelBridge::setLogFileName( const std::string &szLogFileName )
{
    m_channel4.setTraceFilePath( szLogFileName );
}


//#########################################################
//####################### HELPER ##########################
//#########################################################

int ChannelBridge::read( uint8_t *buffer, uint32_t receiveCount, uint32_t &receivedCount )
{
    return m_channel4.readData( buffer, receiveCount, receivedCount );
}

int ChannelBridge::write( const uint8_t *buffer, uint32_t transmitCount, uint32_t &transmittedCount )
{
    return m_channel4.writeData( buffer, transmitCount, transmittedCount );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
uint8_t ChannelBridge::getProtocolVersion( uint8_t deviceNumber )
{
    switch ( m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber] )
    {
        case 1: // Version 1 des Protokolls
        case 2: // Version 2 des MM-Protokolls
        case 3: // Mehrfachdockingstation
        case 4: // Version 4 des MM-Protokolls
            return m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber];

        default: // Ermitteln der Protokollversion

            // Bevorzugt auf AEIII+ pruefen?
            if ( m_prefer_AEIII && (m_channel1_2.Request( (uint8_t)deviceNumber ) == TRUE) )
            {
                // Pruefen ob Protokollversion 1 vorhanden
                return m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber] = 1;
            }
            else if ( isOpen() && m_channel4.protocol().isTimeboySlot( deviceNumber ) )
            {
                // Pruefen ob Protokollversion 4 vorhanden
                return m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber] = 4;
            }
            else if ( isOpen() && (m_channel4.ping( deviceNumber, nullptr, nullptr ) >= IV_ERROR_SUCCESS || m_channel4.lastResult() == -IV_ERROR_NEED_USER_PASSWORD) )
            {
                // Pruefen ob Protokollversion 4 vorhanden
                return m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber] = 4;
            }
            else if ( isOpen() ) // Schnittstelle kann ggf. aufgrund von Fehlern bereits geschlossen sein.
            {
                // Handelt es sich ggf. um eine Dockingstation der Version 1
                if ( m_dockingPreferType == 2 )
                {
                    m_channel4.protocol().setTimeboyStationEnabled( true );
                    if ( m_channel4.protocol().isTimeboySlot( deviceNumber ) )
                        // Pruefen ob Protokollversion 4 vorhanden
                        return m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber] = 4;

                    m_channel4.protocol().setTimeboyStationEnabled( false );
                }

                // Hoher Modbustimeout aelterer Firmwarestaende uebergehen.
                std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );

                if ( m_channel1_2.Request2( (uint8_t)deviceNumber ) == TRUE )
                {
                    // Pruefen ob Protokollversion 2 vorhanden
                    return m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber] = 2;
                }
                else if ( isOpen() ) // Schnittstelle kann ggf. aufgrund von Fehlern bereits geschlossen sein.
                {
                    // Hoher Modbustimeout aelterer Firmwarestaende uebergehen.
                    std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );

                    if ( m_channel1_2.Request( (uint8_t)deviceNumber ) == TRUE )
                    {
                        // Pruefen ob Protokollversion 1 vorhanden
                        return m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber] = 1;
                    }
                }
            }
    }

    return 0;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
uint8_t ChannelBridge::setProtocolVersion( uint8_t deviceNumber, uint8_t version )
{
    uint8_t old = m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber];

    m_protocol_version[(deviceNumber == 0xFE) ? 0 : deviceNumber] = version;

    return old;
}

//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

int ChannelBridge::collectData( uint8_t deviceNumber, int *piError )
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    // Löschen des Datensatzpuffers
    if ( m_recordSize )
    {
        m_recordCollected = false;
        memset( m_recordData, 0, sizeof( m_recordData ) );
        m_recordSize = 0;
        m_recordError = 0;
    }

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
    case 1:
    case 2: {
        m_recordSize = sizeof( m_recordData );

        // Datensatz ermitteln
        iRes = m_channel1_2.ReadRecord(deviceNumberReceiver(), m_recordData, m_recordSize, &m_recordError);
        break;
    }

    case 4: {
        uint8_t status = 0;
        m_recordSize = sizeof( m_recordData );

        // Befehlsausfuehrung starten.
        iRes = m_channel4.readRecord( deviceNumberReceiver(), m_recordData, m_recordSize, status );

        iRes = resultToError(iRes, &m_recordError);
        if (iRes)
        {
            // OK
            iRes = status;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return -1;
    }

    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden	
        m_recordCollected = 0;
        m_recordSize = 0;
        m_recordError = 0;
        memset( m_recordData, 0, sizeof( m_recordData ) );
        return 0;

    case 1: // Online Datensatz
    case 2: // Wiederholter Online Datensatz
        m_recordCollected = true;
        return 2;

    case 3: // Offline Datensatz
    case 4: // Wiederholter Offline Datensatz
        m_recordCollected = true;
        return 1;

    default:
        // Fehler bei Ermittlung
        *piError = m_recordError;
    }

    return -1;
}

int ChannelBridge::getDatensatz( uint8_t *buffer, uint16_t &size, int32_t &error )
{
    int iRes = -1;

    // Wenn Flag für Collected nicht gesetzt, dann wird aktuell kein Datensatz vorgehalten.
    if ( !m_recordCollected )
    {
        return 0; // Es liegen kein Datensatz zum kopieren vor. Erst CollectData aufrufen.
        //error = IV_ERROR_NO_DATA_AVAILABLE;
        //return -1;
    }

    // Zu grosse Datensaetze abfangen!
    if ( m_recordSize > size )
    {
        error = IV_ERROR_FUNCTION_PARAMETER;
        return -1;
    }

    for ( uint16_t i = 0; i < m_recordSize; ++i )
    {
        buffer[i] = m_recordData[i];
    }
    size = m_recordSize;
    // Aktuell kein Datensatz zwischengespeichert.
    m_recordCollected = false;

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
        {
            // Datensatz quittieren (3x -> Versuchen)
            for ( int i = 0; (i < 3) && (iRes = m_channel1_2.QuitRecord( deviceNumberReceiver(), &m_recordError )) == -1; i++ )
            {

            }

            // Quittierung erfolgreich?
            if ( iRes )
            {
                m_recordSize = sizeof( m_recordData );
                // Naechsten Datensatz mit Duplikattest anfordern
                for ( int i = 0; (i < 3) && (iRes = m_channel1_2.ReadRecord( deviceNumberReceiver(), m_recordData, m_recordSize, &m_recordError )); i++ )
                {

                }
            }
            break;
        }
        case 4:
        {
            uint8_t status = 0;
            std::string param;
            // Befehlsausfuehrung starten.
            iRes = m_channel4.quitRecord( deviceNumberReceiver(), status, param );

            iRes = resultToError( iRes, &m_recordError );
            if ( iRes )
            {
                // OK
                // naechsten Datensatz lesen.
                uint8_t status = 0;
                m_recordSize = sizeof( m_recordData );

                // Befehlsausfuehrung starten.
                iRes = m_channel4.readRecord( deviceNumberReceiver(), m_recordData, m_recordSize, status );

                iRes = resultToError( iRes, &m_recordError );
                if ( iRes )
                {
                    // OK
                    iRes = status;
                }
                else
                {
                    iRes = -1;
                }
            }
            else
            {
                iRes = -1;
            }
            break;
        }

        default:
            error = IV_ERROR_UNREACHABLE;
            return -1;
    }

    switch ( iRes )
    {
        case 0: // Keine Datensaetzte vorhanden
            // Aktuell uebergebener ist letzter moeglicher, bei Quittierungsmisserfolg.
            m_recordCollected = false;
            memset( m_recordData, 0, sizeof( m_recordData ) );
            m_recordSize = 0;
            m_recordError = 0;
            return 2;

            // Online Datensaetze
        case 1: // Online Datensatz
        case 2: // Wiederholter Online Datensatz
            m_recordCollected = true;
            return 1;

            // Offline Datensaetze
        case 3: // Offline Datensatz
        case 4: // Wiederholter Offline Datensatz
            m_recordCollected = true;
            return 1;

        default:
            error = m_recordError;
    }

    return -1;
}

int ChannelBridge::systemReadRecord( uint8_t deviceNumber, uint8_t *buffer, uint16_t &size, int32_t &error )
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            error = IV_ERROR_VERSION_FIRMWARE;
            return -1;

        case 4:
        {
   // OK
   // naechsten Datensatz lesen.
            uint8_t status = 0;
            m_recordSize = sizeof( m_recordData );

            // Befehlsausfuehrung starten.
            iRes = m_channel4.readSystemLogRecord( deviceNumberReceiver(), m_recordData, m_recordSize, status );

            iRes = resultToError( iRes, &m_recordError );
            if ( iRes )
            {
                // OK
                iRes = status;
            }
            else
            {
                iRes = -1;
            }
            break;
        }

        default:
            error = IV_ERROR_UNREACHABLE;
            return -1;
    }

    switch ( iRes )
    {
        case 0: // Keine Datensaetze vorhanden
            memset( m_recordData, 0, sizeof( m_recordData ) );
            m_recordSize = 0;
            m_recordError = 0;
            return 0;

        case 1: // System Datensatz
            // Zu grosse Datensaetze abfangen!
            if ( m_recordSize > size )
            {
                error = IV_ERROR_FUNCTION_PARAMETER;
                return -1;
            }

            // Daten in Puffer uebergeben, zu grosse Datensaetze sind abfangen!
            for ( int i = 0; i < m_recordSize; ++i )
            {
                buffer[i] = m_recordData[i];
            }
            size = m_recordSize;
            return iRes;

        default:
            // Fehler bei Ermittlung
            error = m_recordError;
            resetWithError( error );
            break;
    }

    return -1;
}

int ChannelBridge::systemQuitRecord( uint8_t deviceNumber, int *error )
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *error = IV_ERROR_VERSION_FIRMWARE;
            return -1;

        case 4:
        {
   // OK
   // naechsten Datensatz lesen.
            uint8_t status = 0;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.quitSystemLogRecord( deviceNumberReceiver(), status );

            iRes = resultToError( iRes, &m_recordError );
            if ( iRes )
            {
                // OK
                iRes = status;
            }
            else
            {
                iRes = -1;
            }
            break;
        }

        default:
            *error = IV_ERROR_UNREACHABLE;
            return -1;
    }

    switch ( iRes )
    {
        case 0: // Keine Datensaetze vorhanden
            return 0;

        case 1: // Datensatz quitiert
            return 1;

        default:
            // Fehlercode uebernehmen
            *error = m_recordError;
    }

    return -1;
}

int ChannelBridge::systemRestoreRecords( uint8_t deviceNumber, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            return 0;

        case 4:
        { // PZEMasterIV mit Protokollversion 4
            int iRes;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.restoreSystemLogRecords( deviceNumberReceiver() );

            return resultToError( iRes, piError );
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::readRecord( uint8_t deviceNumber, uint8_t *buffer, uint16_t &size, int32_t &error )
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
        {
            m_recordSize = sizeof( m_recordData );
            // Datensatz ermitteln
            iRes = m_channel1_2.ReadRecord( deviceNumberReceiver(), m_recordData, m_recordSize, &m_recordError );
            break;
        }

        case 4:
        {
   // OK
   // naechsten Datensatz lesen.
            uint8_t status = 0;
            m_recordSize = sizeof( m_recordData );

            // Befehlsausfuehrung starten.
            iRes = m_channel4.readRecord( deviceNumberReceiver(), m_recordData, m_recordSize, status );

            iRes = resultToError( iRes, &m_recordError );
            if ( iRes )
            {
                // OK
                iRes = status;
            }
            else
            {
                iRes = -1;
            }
            break;
        }

        default:
            error = IV_ERROR_UNREACHABLE;
            return -1;
    }

    switch ( iRes )
    {
        case 0: // Keine Datensaetze vorhanden
            memset( m_recordData, 0, sizeof( m_recordData ) );
            m_recordSize = 0;
            m_recordError = 0;
            return 0;

        case 1: // Online Datensatz
        case 2: // Wiederholter Online Datensatz
        case 3: // Offline Datensatz
        case 4: // Wiederholter Offline Datensatz

            // Daten in Puffer uebergeben
            // Zu grosse Datensaetze abfangen!
            if ( m_recordSize > size )
            {
                error = IV_ERROR_FUNCTION_PARAMETER;
                return -1;
            }
            for ( int i = 0; i < m_recordSize; ++i )
            {
                buffer[i] = m_recordData[i];
            }

            size = m_recordSize;
            return iRes;

        default:
            // Fehler bei Ermittlung
            error = m_recordError;
            resetWithError( error );
            break;
    }

    return -1;
}

int ChannelBridge::quitRecord( uint8_t deviceNumber, std::string &param, int *error )
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:

            // Datensatz ermitteln
            iRes = m_channel1_2.QuitRecord( deviceNumberReceiver(), &m_recordError );
            break;

        case 4:
        {
   // OK
   // naechsten Datensatz lesen.
            uint8_t status = 0;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.quitRecord( deviceNumberReceiver(), status, param );

            iRes = resultToError( iRes, &m_recordError );
            if ( iRes )
            {
                // OK
                iRes = status;
            }
            else
            {
                iRes = -1;
            }
            break;
        }

        default:
            *error = IV_ERROR_UNREACHABLE;
            return -1;
    }

    switch ( iRes )
    {
        case 0: // Keine Quittung noetig	
            return 0;

        case 1: // Online Datensatzquit.
        case 2: // Wiederholte Online Datensatzquit.
        case 3: // Offline Datensatzquit.
        case 4: // Wiederholte Offline Datensatzquit.
            return iRes;

        default:
            // Fehler bei Ermittlung
            *error = m_recordError;
            resetWithError( *error );
            break;
    }

    return -1;
}

// Loescht alle Datensaetze
int ChannelBridge::clearData( uint8_t deviceNumber, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.ClearData( deviceNumberReceiver(), piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4
            int iRes;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.clearRecords( deviceNumberReceiver() );

            return resultToError( iRes, piError );
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Stellt den Datensatzzeiger an Ringbeginn. Alle Datensaetze wieder abrufbar.
int ChannelBridge::restoreRecords( uint8_t deviceNumber, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            return 0;

        case 4:
        { // PZEMasterIV mit Protokollversion 4
            int iRes;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.restoreRecords( deviceNumberReceiver() );

            return resultToError( iRes, piError );
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getDateTime( uint8_t deviceNumber, uint8_t *pucDateTime, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:

            if ( !m_channel1_2.ReadMem( MB_GET_TIME, 7, pucDateTime, piError ) )
            {
                return 0;
            }
            return 1;

        case 4:
        {
            int iRes = IV_ERROR_SUCCESS;
            char szTime[20];

            // Befehlsausfuehrung starten.
            iRes = m_channel4.readRtcDateTime( deviceNumberReceiver(), szTime );

            iRes = resultToError( iRes, piError );
            if ( iRes )
            {
                // OK
                pucDateTime[0] = (szTime[0] - 0x30) * 10 + szTime[1] - 0x30;
                pucDateTime[1] = (szTime[2] - 0x30) * 10 + szTime[3] - 0x30;

                pucDateTime[2] = (szTime[5] - 0x30) * 10 + szTime[6] - 0x30;

                pucDateTime[3] = (szTime[8] - 0x30) * 10 + szTime[9] - 0x30;

                pucDateTime[4] = (szTime[11] - 0x30) * 10 + szTime[12] - 0x30;
                pucDateTime[5] = (szTime[14] - 0x30) * 10 + szTime[15] - 0x30;
                pucDateTime[6] = (szTime[17] - 0x30) * 10 + szTime[18] - 0x30;
            }

            return iRes;
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::setDateTime( uint8_t deviceNumber, const uint8_t *pucDateTime, int *piError )
{
    uint8_t ucDateTime[8];

    if ( pucDateTime == NULL ||
        (pucDateTime[0] != 20) ||
         (pucDateTime[1] > 99) ||
         (pucDateTime[2] < 1) || (pucDateTime[2] > 12) ||
         (pucDateTime[3] < 1) || (pucDateTime[3] > 31) ||
         (pucDateTime[4] > 23) || (pucDateTime[5] > 59) || (pucDateTime[6] > 59) )
    {
        time_t tm_time;
        struct tm *ptm;
        tm_time = time( NULL );
        ptm = localtime( &tm_time );

        ucDateTime[0] = (ptm->tm_year + 1900) / 100;
        ucDateTime[1] = (ptm->tm_year + 1900) % 100;
        ucDateTime[2] = ptm->tm_mon + 1;
        ucDateTime[3] = ptm->tm_mday;
        ucDateTime[4] = ptm->tm_hour;
        ucDateTime[5] = ptm->tm_min;
        ucDateTime[6] = ptm->tm_sec;
    }
    else
    {
        memcpy( ucDateTime, pucDateTime, 7 );
    }

    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    int iRes = IV_ERROR_SUCCESS;
    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:

            if ( !m_channel1_2.WriteMem( MB_SET_TIME, 7, ucDateTime, piError ) )
            {
                return 0;
            }
            return 1;

        case 4:
        {
            char szTime[28];
            snprintf( szTime, sizeof( szTime ) - 1, "%02i%02i-%02i-%02i %02i:%02i:%02i",
                      ucDateTime[0], ucDateTime[1], ucDateTime[2], ucDateTime[3],
                      ucDateTime[4], ucDateTime[5], ucDateTime[6] );

                  // Befehlsausfuehrung starten.
            iRes = m_channel4.writeRtcDateTime( deviceNumberReceiver(), szTime );

            return resultToError( iRes, piError );
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::sendInfotext( uint8_t deviceNumber, const std::string &text, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        {
            uint8_t Buffer[245];
            uint32_t counter;
            uint32_t sizeText = static_cast<uint32_t>(text.size());

            Buffer[0] = 0xFF;
            Buffer[1] = 0;
            Buffer[2] = 0;

            for ( counter = 0; counter < sizeText; counter++ )
            {
                Buffer[counter + 3] = text[counter];
            }

            if ( !m_channel1_2.WriteMem( INFOTEXT_MEMORY, sizeText + 3, &Buffer[0], piError ) )
            {
                return 0;
            }

            return 1;
        }

        case 2:
            *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
            return 0;

        case 4:
        {
            int iRes = IV_ERROR_SUCCESS;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.writeMemoToDisplay( deviceNumberReceiver(), 0, 0, 0, 0, 0, text );

            return resultToError( iRes, piError );
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::sendMessage( uint8_t deviceNumber, uint8_t ucDelayTime,
                                uint8_t ucAction, uint8_t ucAudio,
                                const std::string &text, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
        {
            uint8_t Buffer[245];
            uint32_t counter;
            uint32_t sizeText = static_cast<uint32_t>(text.size());

            Buffer[0] = ucDelayTime;
            Buffer[1] = ucAction;
            Buffer[2] = ucAudio;

            for ( counter = 0; (counter < sizeText) && (counter < (int)(sizeof( Buffer ) - 3)); counter++ )
            {
                Buffer[3 + counter] = text[counter];
            }

            if ( !m_channel1_2.WriteMem( DISPLAY_MEMORY, counter + 3, &Buffer[0], piError ) )
            {
                return 0;
            }
            return 1;
        }

        case 4:
        {
            int iRes = IV_ERROR_SUCCESS;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.writeMessageToDisplay( deviceNumberReceiver(), ucDelayTime, ucAction, ucAudio, text );

            return resultToError( iRes, piError );
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Datensatzbeschreibung oder Listenbeschreibung aus dem Geraet lesen.
int ChannelBridge::getTableDescription( uint8_t deviceNumber, uint8_t ucList, int *piError )
{
    int res = 0;
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    // Für den späteren Zugriff auf die Daten des Puffers die ProtocolVersion halten.
    uint8_t protocolVersion = getProtocolVersion( deviceNumber );

    switch ( protocolVersion )
    {
        case 1:  // AEIII+
        case 2:  // PZE - Master IV        
            res = m_channel1_2.GetTables( ucList, piError );
            break;

        case 4:
            // Befehlsausfuehrung starten.
            res = resultToError( m_channel4.readTablesWithTimeboySlot( deviceNumber, (ucList) ? (char *)"LIST" : (char *)"RECORD" ), piError );
            break;

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    if ( res )
    {
        if ( ucList )
            m_protocol_version_list_description = protocolVersion;
        else
            m_protocol_version_record_description = protocolVersion;
    }

    return res;
}

// Listenbeschreibung aus dem Geraet lesen.
int ChannelBridge::getListName( uint8_t deviceNumber, uint8_t ucList, std::string &name, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.GetListName( ucList, name, piError );
            break;

        case 4:
            *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
            return 0;

        default:
            *piError = IV_ERROR_UNREACHABLE;
            return 0;
    }

    return 0;
}

int ChannelBridge::writeStudioFile( uint8_t deviceNumber, const std::string &filePath, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.WriteSetupData( filePath, piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4
            int iRes = IV_ERROR_SUCCESS;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.writeSetupWithTimeboySlot( deviceNumberReceiver(), filePath );

            return resultToError( iRes, piError );
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::readStudioFile( uint8_t deviceNumber, const std::string &filePath, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    int res = 0;
    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.ReadSetupData( filePath, piError );

        case 4: // PZEMasterIV mit Protokollversion 4
            return resultToError( m_channel4.readSetupWithTimeboySlot( deviceNumberReceiver(), filePath ), piError );

        default:
            *piError = IV_ERROR_UNREACHABLE;
            res = 0;
    }

    return res;
}

int ChannelBridge::modifyStudioFile( const std::string &sourceFile, const std::string &destFile, const std::string &changeData, int flags, int *piError )
{
    (void)sourceFile;
    (void)destFile;
    (void)changeData;
    (void)flags;
    (void)piError;

#if !defined(DATAFOX_STUDIO_IV)
    CDFFileStudioXML file;
    int fileError = 0;
    struct stat st;

    if ( stat( sourceFile.c_str(), &st ) != 0 )
    {
        // Datei konnte nicht ermittelt werden
        *piError = IV_ERROR_FILE_OPEN;
        return 0;
    }

    // Reset der StudioFile-Klasse
    file.SetDefault();
    file.SetLibraryVersion( g_dfcomVersion );

    // Laden der Dateidaten. Keine Beachtung der Studioversion.
    if ( !file.load( sourceFile, &fileError, static_cast<FileAES::Enum_flag_ignore_version>(FileAES::fiv_min_studio | FileAES::fiv_should_studio) ) )
    {
        *piError = fileErrorToErrorCode( fileError );
        return 0;
    }

    // StudioDatei modifizieren
    if ( file.Modify( changeData.c_str(), piError ) )
    {
        // StudioDatei speichern
        if ( file.save( destFile, &fileError ) )
        {
            // Alles erfolgreich durchgefuehrt
            return 1;
        }
        else
        {
            // Datei konnte nicht erstellt werden.
            *piError = IV_ERROR_FILE_OPEN;
        }
    }
#endif
    return 0;
}

int ChannelBridge::convertFile( const std::string &srcFile, FileType srcType, const std::string &dstFile, FileType dstType, int flags, int *piError )
{
    // Checks the convertibility of source type or file extension of the source to target type or file extension of the target.
    // The file extension is only used if the file type was not explicitly specified (FileType::Undefined).
    FileType tmpSrcType = (srcType == FileType::Undefined) ? getFileTypeByExtension( srcFile ) : srcType;
    FileType tmpDstType = (dstType == FileType::Undefined) ? getFileTypeByExtension( dstFile ) : dstType;

    if ( !isConvertible( tmpSrcType, tmpDstType ) )
    {
        // Conversion not available / implemented
        *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
        return 2;
    }

    std::unique_ptr<DocumentStudio> pDoc( new DocumentStudio( "", std::string( g_dfcomVersion ) ) );
    pDoc->fileXML().configureAdditionalHelpFile( m_createXMLHelpFile, m_docPath );

    switch ( pDoc->documentOpen( srcFile, true ).code )
    {
        case DocumentStudio::ResultCodes::Success:
        {
            *piError = pDoc->documentSaveAs( tmpDstType, dstFile ) ? IV_ERROR_SUCCESS : IV_ERROR_FILE_WRITE;
            break;
        }

        case DocumentStudio::ResultCodes::OpenUnsupportedFileType:
        case DocumentStudio::ResultCodes::OpenUnsupportedFileVersion:
        {
            *piError = IV_ERROR_FILE_EXTENSION;
            break;
        }

        case DocumentStudio::ResultCodes::FileContentCorrupted:
        case DocumentStudio::ResultCodes::SetupOpeningFailed:
        case DocumentStudio::ResultCodes::BackupCreationFailed:
        case DocumentStudio::ResultCodes::FileAES_openFailed:
        {
            *piError = IV_ERROR_FILE_OPEN;
            break;
        }

        case DocumentStudio::ResultCodes::FileReadingError:
        case DocumentStudio::ResultCodes::FileAES_readFailed:
        {
            *piError = IV_ERROR_FILE_READ;
            break;
        }

        case DocumentStudio::ResultCodes::FileLoadingError:
        case DocumentStudio::ResultCodes::TooMuchDataError:
        case DocumentStudio::ResultCodes::FileAES_versionTooNew:
        case DocumentStudio::ResultCodes::FileAES_recommendedVersionError:
        case DocumentStudio::ResultCodes::FileAES_DFComDllToOld:
        case DocumentStudio::ResultCodes::FileAES_wrongFormat:
        case DocumentStudio::ResultCodes::FileXML_structureError:
        {
            *piError = IV_ERROR_FILE_FORMAT;
            break;
        }

        case DocumentStudio::ResultCodes::FileAES_checksumError:
        case DocumentStudio::ResultCodes::FileAES_contentChecksumError:
        {
            *piError = IV_ERROR_FILE_CRC;
            break;
        }

        case DocumentStudio::ResultCodes::FileAES_internalError:
        case DocumentStudio::ResultCodes::FileXML_internalError:
        {
            *piError = IV_ERROR_INTERN;
            break;
        }
    }

    return (*piError == 0) ? 1 : 0;
}

int ChannelBridge::fileErrorToErrorCode( const int nFileError ) const
{
    switch ( nFileError )
    {
        case -3:
        {
            return IV_ERROR_FILE_READ;
        }
        case -7: // Minimum version of the DLL is not adhered to.
        {
            return IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
        }
        case -8: // The recommended version of the DLL is not available.
        {
            return IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION;
        }
        case -9: // Incorrect format of the setup file
        {
            return IV_ERROR_FILE_FORMAT;
        }
        case -4: // CRC error in the file data
        case -10: case -11: case -12: case -13: case -14: case -15: case -16: case -17: case -18:
        {
            return IV_ERROR_FILE_FORMAT;
        }
    }
    return IV_ERROR_FILE_OPEN;
}

bool ChannelBridge::isConvertible( FileType srcType, FileType dstType ) const
{
    // Conversion is not possible if the source or destination file type is not defined.
    if ( (srcType == FileType::Undefined) || (dstType == FileType::Undefined) )
    {
        return false;
    }

    // The Source and destination type are equal. No conversion necessary.
    if ( srcType == dstType )
    {
        return false;
    }

    bool bResult = false;
    switch ( srcType )
    {
        case FileType::AES_AEL:
        {
            bResult = (dstType == FileType::XML);
            break;
        }
        case FileType::XML:
        {
            bResult = (dstType == FileType::AES_AEL);
            break;
        }
    }
    return bResult;
}

FileType ChannelBridge::getFileTypeByExtension( const std::string &file ) const
{
    FileType result = FileType::Undefined;

    std::string ext( file.substr( file.size() - 4 ) );
    boost::to_lower( ext );

    if ( ext == ".aes" || ext == ".ael" )
    {
        result = FileType::AES_AEL;
    }
    else if ( ext == ".xml" )
    {
        result = FileType::XML;
    }

    return result;
}

int ChannelBridge::checkDevice( uint8_t deviceNumber, int *piError, int iPollRetry )
{
    int iRes = IV_ERROR_SUCCESS;

    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.CheckDevice( iPollRetry, piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4
            std::string version;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.readFirmwareVersion( deviceNumberReceiver(), version );

            return resultToError( iRes, piError );
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return 0;
}

int ChannelBridge::restart( uint8_t deviceNumber, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 2:
            return m_channel1_2.Restart( piError );

        case 1:
        case 4:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            break;

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return 0;
}

int ChannelBridge::reset( uint8_t deviceNumber, int mode, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    int iRes = IV_ERROR_SUCCESS;
    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            break;

        case 4:
            // Befehlsausfuehrung starten.
            iRes = m_channel4.systemReset( deviceNumber, mode );

            iRes = resultToError( iRes, piError );
            break;

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return iRes;
}

int ChannelBridge::getProtocolVersion2( uint8_t deviceNumber, uint8_t detailed, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    uint8_t Buf[4];

    // Protokollverwendung pruefen
    uint8_t pv = getProtocolVersion( deviceNumberReceiver() );
    switch ( pv )
    {
        case 1:
        case 2:
            // Detailierte Version auslesen.
            if ( detailed )
            {
                // Version fuer Setup lesen
                if ( !m_channel1_2.ReadMem( MB_GET_VERSION, 2, Buf, piError ) )
                {
                    return 0;
                }
            }
            else
            {
                if ( pv == 1 )
                {
                    // Es wird angenommen ein AEIII+ mit neuester Com-Version
                    Buf[0] = 0x04;
                    Buf[1] = 0x02;
                }
                else
                {
                    // Es wird angenommen ein PZE Master IV mit neuester Com-Version
                    Buf[0] = 0x05;
                    Buf[1] = 0x01;
                }
            }
            break;

        case 3:
            // Ab dem Protokolltyp 3 wird nur noch ueber die Firmwareversion entschieden.
            Buf[0] = 0x06;
            Buf[1] = 0x00;
            break;

        case 4:
            // Ab dem Protokolltyp 4 wird nur noch ueber die Firmwareversion entschieden.
            Buf[0] = 0x07;
            Buf[1] = 0x00;
            break;

        default:
            *piError = IV_ERROR_UNREACHABLE;
            return 0;
    }


    // Version fuer Kommunikation

    // Ab Firmware 04.01.00 Protokollanpassung (Version 4) und Listen in Tabellenart abhandeln.
    if ( (Buf[0] == 7 && Buf[1] == 0) ||
        // Ab Firmware 04.01.00 Protokollanpassung (Version 3)
        (Buf[0] == 6 && Buf[1] == 0) ||
        // Ab Firmware 04.00.23 Listengroesse auf 256kByte erhoeht.
         (Buf[0] == 5 && Buf[1] == 2) ||
         // Ab Firmware 04.00.21 Listen im Hintergrund laden
         (Buf[0] == 5 && Buf[1] == 1) ||
         // Ab Firmware 04.00.19 des PZEMasterIV (Version 2)
         (Buf[0] == 5 && Buf[1] == 0) ||
         // Ab Firmware 03.01.21 wegen auslesen der Firmwareversion
         (Buf[0] == 4 && Buf[1] == 2) ||
         // Ab Firmware 03.00.12 mit Zutrittskontrolle
         (Buf[0] == 4 && Buf[1] == 1) ||
         // Ab Firmware 3.0 mit 16kByte Flash fuer Setupdaten und 16kByte Extended Flash
         (Buf[0] == 4 && Buf[1] == 0) )
    {
        // Version als Rueckgabe
        return (Buf[0] << 8) | Buf[1];
    }

    *piError = IV_ERROR_VERSION_FIRMWARE;
    return 0;
}

int ChannelBridge::getSerialNum( uint8_t deviceNumber, int *piSerialNum, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.GetSerialNum( piSerialNum, piError );

        case 4:
        {
            int iRes;
            uint32_t serial;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.readSerialNumber( deviceNumberReceiver(), serial );

            iRes = resultToError( iRes, piError );
            if ( iRes )
            {
                // OK
                *piSerialNum = serial;
            }

            return iRes;
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeListData( uint8_t deviceNumber, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.WriteListData( &m_channel4, piError );

        case 4: // PZEMasterIV mit Protokollversion 4

            // Befehlsausfuehrung starten.
            return resultToError( m_channel4.writeListWithTimeboySlot( deviceNumberReceiver() ), piError );

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeEntranceListData( uint8_t deviceNumber, int iListNum, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    // Protokollklasse
    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 4:
            *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
            return 0;

        case 1:
        case 2:
            return m_channel1_2.WriteEntranceListData( iListNum, piError );

        default:
            *piError = IV_ERROR_UNREACHABLE;
            return 0;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeEntrance2ListData( uint8_t deviceNumber, int iListNum, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            return 0;

        case 4:
            // Befehlsausfuehrung starten.
            // Listendaten zum AEIV uebertragen
            return resultToError( m_channel4.writeEntrance2ListWithTimeboySlot( deviceNumber, iListNum ), piError );

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeTimeboyListData( uint8_t deviceNumber, uint8_t id, uint8_t num, int reserve, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            return 0;

        case 4:
            // Befehlsausfuehrung starten.
            // Listendaten zum AEIV uebertragen
            return resultToError( m_channel4.writeTimeboyListWithTimeboySlot( deviceNumber, id, num, reserve ), piError );

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int ChannelBridge::checkData( uint8_t deviceNumber, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:  // AEIII+
        case 2:  // PZEMasterIV
            return m_channel1_2.CheckData( piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4
            int iRes = IV_ERROR_SUCCESS;
            uint8_t buf[0x255];
            uint16_t count = sizeof( buf );
            uint8_t status = 0;

            // Befehlsausfuehrung starten.
            iRes = m_channel4.readRecord( deviceNumberReceiver(), buf, count, status );

            iRes = resultToError( iRes, piError );
            if ( iRes )
            {
                // OK
                iRes = status;
            }
            else
            {
                iRes = -1;
            }
            return iRes;
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return -1;
}

// Schliesst Relais mit Nummer fuer Zeit.
int ChannelBridge::closeRelay( uint8_t deviceNumber, int iNum, int iTime, int *piError )
{
    int iRes = IV_ERROR_SUCCESS;
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.CloseRelay( deviceNumberReceiver(), iNum, iTime, piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4
//         if (iNum > 4)
//         {
//             *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
//             iRes = 0;
//         }
//         else
//         {
            // Befehlsausfuehrung starten.
            iRes = resultToError( m_channel4.writeRelayState( deviceNumberReceiver(), (char *)"INTERN", 0, iNum, 1, iTime ), piError );
//         }
            return iRes;
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Ermittelt aktuellen Zustand von Relais mit Nummer fuer Zeit.
int ChannelBridge::getRelayState( uint8_t deviceNumber, int iNum, int *piState, int *piTime, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.GetRelayState( deviceNumberReceiver(), iNum, piState, piTime, piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4
            int iRes = IV_ERROR_SUCCESS;
    //         if (iNum > 4)
    //         {
    //             *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
    //             iRes = 0;
    //         }
    //         else
    //         {		
            uint8_t state;
            uint16_t time;

            // Befehlsausfuehrung starten.
            iRes = resultToError( m_channel4.readRelayState( deviceNumberReceiver(), (char *)"INTERN", 0, iNum, state, time ), piError );

            if ( iRes )
            {
                // OK
                *piState = state;
                *piTime = time;
            }
//        }
            return iRes;
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Oeffnet Relais mit Nummer.
int ChannelBridge::openRelay( uint8_t deviceNumber, int iNum, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    int iRes = IV_ERROR_SUCCESS;

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.OpenRelay( deviceNumberReceiver(), iNum, piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4
//         if (iNum > 4)
//         {
//             *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
//             iRes = 0;
//         }
//         else
//         {		
            uint16_t time = 0;

            // Befehlsausfuehrung starten.
            iRes = resultToError( m_channel4.writeRelayState( deviceNumberReceiver(), (char *)"INTERN", 0, iNum, 0, time ), piError );
//         }
            return iRes;
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Daten einer Variablen setzen.
int ChannelBridge::setGlobVar( uint8_t deviceNumber, const char *name, int iType, const char *val, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    int iRes = IV_ERROR_SUCCESS;

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.SetGlobVar( deviceNumberReceiver(), name, iType, val, piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4

   // Befehlsausfuehrung starten.
            switch ( iType )
            {
                case 0:
                {
                    char varName[64];
                    memset( varName, 0, sizeof( varName ) );
                    snprintf( varName, sizeof( varName ), "SETUP.%i", *(uint8_t *)name );
                    iRes = m_channel4.writeSetupVariable( deviceNumberReceiver(), varName, val );
                    break;
                }

                case 1:
                {
                    char varName[64];
                    memset( varName, 0, sizeof( varName ) );
                    snprintf( varName, sizeof( varName ), "SETUP.%s", name );
                    iRes = m_channel4.writeSetupVariable( deviceNumberReceiver(), varName, val );
                    break;
                }

                case 2:
                    iRes = m_channel4.writeSystemVariable( deviceNumberReceiver(), name, val );
                    break;
            }

            if ( iRes > 0 )
            {
                switch ( iRes )
                {
                    case 1:
                    case 2: // Namensbereichsangabe oder Variablenname unbekannt.
                        *piError = IV_ERROR_VARIABLE_UNDEFINED;
                        iRes = 0;
                        break;

                    case 3: // Der verwendete Variablenwert ist zu lang.
                        *piError = IV_ERROR_VARIABLE_SIZE;
                        iRes = 0;
                        break;

                    case 4: // Der verwendete Variablentype würde nicht eingehalten.
                        *piError = IV_ERROR_VARIABLE_TYPE;
                        iRes = 0;
                        break;

                    case 5: // Das Schreib-/Leserecht würde nicht eingehalten.
                        *piError = IV_ERROR_VARIABLE_RIGHT;
                        iRes = 0;
                        break;

                    case 6: // Das verwendete Format oder der Wertebereich würde nicht eingehalten.
                        *piError = IV_ERROR_VARIABLE_FORMAT_OR_RANGE;
                        iRes = 0;
                        break;

                    default:
                        iRes = resultToError( iRes, piError );
                }
            }
            else
            {
                iRes = resultToError( iRes, piError );
            }
            return iRes;
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Daten einer Globalen Variablen lesen.
int ChannelBridge::getGlobVar( uint8_t deviceNumber, const char *name, int iType, char *val, int iValueLength, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    // muss mindestens ein Zeichen anfordern.
    if ( (iValueLength - 1) <= 0 )
    {
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return 0;
    }

    int iRes = IV_ERROR_SUCCESS;
    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            return m_channel1_2.GetGlobVar( deviceNumberReceiver(), name, iType, val, iValueLength, piError );

        case 4:
        { // PZEMasterIV mit Protokollversion 4

            std::string lValue;
            // Befehlsausfuehrung starten.
            switch ( iType )
            {
                case 0:
                {
                    char varName[64];
                    memset( varName, 0, sizeof( varName ) );
                    snprintf( varName, sizeof( varName ), "SETUP.%i", *(uint8_t *)name );
                    iRes = m_channel4.readSetupVariable( deviceNumber, varName, lValue );
                    break;
                }

                case 1:
                {
                    char varName[64];
                    memset( varName, 0, sizeof( varName ) );
                    snprintf( varName, sizeof( varName ), "SETUP.%s", name );
                    iRes = m_channel4.readSetupVariable( deviceNumber, varName, lValue );
                    break;
                }

                case 2:
                    iRes = m_channel4.readSystemVariable( deviceNumber, name, lValue );
                    break;
            }

            if ( iRes > 0 )
            {
                switch ( iRes )
                {
                    case 1:
                    case 2: // Namensbereichsangabe oder Variablenname unbekannt.
                        *piError = IV_ERROR_VARIABLE_UNDEFINED;
                        iRes = 0;
                        break;

                    case 3: // Der verwendete Variablenwert ist zu lang.
                        *piError = IV_ERROR_VARIABLE_SIZE;
                        iRes = 0;
                        break;

                    case 4: // Der verwendete Variablentype würde nicht eingehalten.
                        *piError = IV_ERROR_VARIABLE_TYPE;
                        iRes = 0;
                        break;

                    case 5: // Das Schreib-/Leserecht würde nicht eingehalten.
                        *piError = IV_ERROR_VARIABLE_RIGHT;
                        iRes = 0;
                        break;

                    case 6: // Das verwendete Format oder der Wertebereich würde nicht eingehalten.
                        *piError = IV_ERROR_VARIABLE_FORMAT_OR_RANGE;
                        iRes = 0;
                        break;

                    default:
                        iRes = resultToError( iRes, piError );
                }
            }
            else
            {
                iRes = resultToError( iRes, piError );
            }
            if ( iRes )
            {
                strncpy( val, lValue.c_str(), iValueLength - 1 );
            }

            return iRes;
        }

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getVersionFirmwareFromFile( const std::string &filePath, std::string &versionFirmware, int *error )
{
    return resultToError( m_channel4.getVersionFirmwareFromFile( filePath, versionFirmware ), error );
}

int ChannelBridge::writeTextData( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.setTextDataWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeTouchKeyData( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.setTouchKeyDataWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeDisplayColors( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.setDisplayColorsWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeWlanConfigData( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.setWlanConfigDataWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readWlanConfigData( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.getWlanConfigDataWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeHardwareConfigData( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.setHardwareConfigDataWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readHardwareConfigData( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.getHardwareConfigDataWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeFileData( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeFileDataWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readFileData( uint8_t deviceNumber, const std::string &filePath )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readFileDataWithTimeboySlot( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeFileDevice( uint8_t deviceNumber, std::string &filePath, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    int iRes = 0;
    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 2: // Protokollversion 2 gibt es nur beim PZE-MasterIV
            iRes = m_channel1_2.WriteFileDevice( filePath, piError );
            break;

        case 4: // MasterIV mit Protokollversion 4

            // Befehlsausfuehrung starten.
            iRes = resultToError( m_channel4.writeDeviceWithTimeboySlot( deviceNumber, filePath ), piError );
            break;
    }

    if ( iRes )
    {
        // Protokollversion ist aufgrund des Updates auf 0 zu setzen.
        // Naechste Kommunikationsanfage ermittelt die Protokollversion neu.
        m_protocol_version[(deviceNumberReceiver() == 0xFE) ? 0 : deviceNumberReceiver()] = 0;
    }

    return iRes;
}

int ChannelBridge::writeModule( uint8_t deviceNumber, int type, const std::string &filePath, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return resultToError( m_channel4.writeModuleWithTimeboySlot( deviceNumber, type, filePath ), piError );
    }

    *piError = IV_ERROR_UNREACHABLE;    // Zeichensatz- und Grafikdaten
    return 0;
}

int ChannelBridge::read_data_block( uint8_t deviceNumber, const std::string &name, char *data, uint32_t *length, int flags )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readDataBlockWithTimeboySlot( deviceNumber, name, data, length, flags );
    }

    return -IV_ERROR_UNREACHABLE;
}

//#########################################################
//######## Schnittstelle zu Channel1/Channel2 #############
//#########################################################

int ChannelBridge::getModus( uint8_t deviceNumber, int *error )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 4: return -IV_ERROR_VERSION_FIRMWARE;
        case 2: return m_channel1_2.GetModus( error );
    }

    return -IV_ERROR_UNREACHABLE;
}


//#########################################################
//############# Schnittstelle zu Channel4 #################
//#########################################################

void ChannelBridge::setCallBackFunctionWithParams( int32_t (*pFunction)(uint32_t, uint32_t) )
{
    m_channel4.setCallBackFunctionWithParameters( pFunction == nullptr ? std::function<int32_t( uint32_t, uint32_t )>() : pFunction );
}

void ChannelBridge::setCallBackBlockTransfer( int32_t (*pFunction)(int32_t, int32_t, uint32_t, uint32_t, uint32_t, uint32_t) )
{
    m_channel4.setCallBackBlockTransfer( pFunction == nullptr ? std::function<int32_t( int32_t, int32_t, uint32_t, uint32_t, uint32_t, uint32_t )>() : pFunction );
}

uint32_t ChannelBridge::crc32( uint8_t *data, uint32_t size, uint32_t crc, bool closure )
{
    return m_channel4.crc32( data, size, crc, closure );
}

// MODULE_HARDWARE

// Einfache Anfrage ob das Geraet antwortet.
int ChannelBridge::ping( uint8_t deviceNumber )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.ping( deviceNumber, nullptr, nullptr );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDevice( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readDeviceType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeBoard( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readBoardType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentRS485_0( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readRs485CommunicationComponent( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTCPIP( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readNetworkAdapterType( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeModemMC( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readModemType( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentRS485_1( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readRs485AccessControlComponent( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDisplay( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readDisplayType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeKeypad( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readKeypadType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTransponder( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readTransponderReaderType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeMMC( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readMmcType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeRadio( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readRadioType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeGPS( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readGpsType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentUSB( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readUsbComponent( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDCF77( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readDCF77Type( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentInAnalog( uint8_t deviceNumber, uint32_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readAnalogInputComponent( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentInDigital( uint8_t deviceNumber, uint32_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readDigitalInputComponent( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentOutDigital( uint8_t deviceNumber, uint32_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readDigitalOutputComponent( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeFingerprint( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readFingerprintScannerType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeExternalReader( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readExternalTransponderReaderType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTouch( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readTouchscreenType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeSensor( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readSensorType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentBattery( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readBatteryComponent( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentCharge( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readBatteryChargerComponent( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionIOModule( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readIoModuleOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionTBAdaption( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readTimeboyAdaptionOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionComServer( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readComServerOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionTPWrite( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readTransponderWriteOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionHTTP( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readHttpOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionLanguage( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readLocalizationOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionBiokeyUpdate( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readFingerprintScannerUpdateOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setSetupOptions( uint8_t deviceNumber, uint16_t mask, uint16_t options )
{
    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeSetupOptions( deviceNumber, mask, options );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSetupOptions( uint8_t deviceNumber, uint16_t &mask, uint16_t &options )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readSetupOptions( deviceNumber, mask, options );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getPayableOptions( uint8_t deviceNumber, uint8_t &mask, uint8_t &options )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readBuyOptions( deviceNumber, mask, options );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHWConfig( uint8_t deviceNumber, uint8_t &val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readHardwareInfoVersion( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHardwareInfo( uint8_t deviceNumber, StructHardwareInfo &hwInfo )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.ModuleTable::readHardwareInfo( deviceNumber, hwInfo );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHardwareInfo( uint8_t deviceNumber, HardwareInfo &hardwareInfo )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readHardwareInfo( deviceNumber, hardwareInfo );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getInfos( StructHardwareInfo hwInfo, std::string &info )
{
    return m_channel4.hardwareInfoToString( hwInfo, info );
}

// MODULE_SYSTEM

// Systemreset; Speicherbereiche loeschen
//int systemReset(uint8_t deviceNumber, uint16_t modus);

// Firmwareversion lesen
int ChannelBridge::getVersionFirmware( uint8_t deviceNumber, std::string &version )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2:
        {
            int errorNumber = IV_ERROR_SUCCESS;
            if ( m_channel1_2.GetVersionFirmware( version, &errorNumber ) )
                return IV_ERROR_SUCCESS;

            return -errorNumber;
        }
        case 4: return m_channel4.readFirmwareVersion( deviceNumber, version );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getVersionBootloader( uint8_t deviceNumber, std::string &version )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1: return -IV_ERROR_VERSION_FIRMWARE;
        case 2:
        {
            int errorNumber = IV_ERROR_SUCCESS;
            if ( m_channel1_2.GetVersionBootloader( version, &errorNumber ) )
                return IV_ERROR_SUCCESS;

            return -errorNumber;
        }
        case 4: return m_channel4.readBootloaderVersion( deviceNumber, version );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSerialNumber( uint8_t deviceNumber, uint32_t &serial )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readSerialNumber( deviceNumber, serial );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::openComServerMode( uint8_t deviceNumber, const char *comm, uint32_t baudRate, uint8_t byteSize, uint8_t parity, uint8_t stopBits, uint16_t flags, uint16_t timeoutToClose )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2:return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.enterComServerMode( deviceNumber, comm, baudRate, byteSize, parity, stopBits, flags, timeoutToClose );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::closeComServerMode( uint8_t deviceNumber )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    return m_channel4.exitComServerMode( deviceNumber );
}

// Datensegmentuebertragung beginnen/enden
int ChannelBridge::beginSegment( uint8_t deviceNumber, uint16_t art, uint32_t &address, uint32_t size, uint32_t crc )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.enterWriteDataSegment( deviceNumber, art, address, size, crc );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::endSegment( uint8_t deviceNumber, uint16_t art, uint32_t crc )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.exitWriteDataSegment( deviceNumber, art, crc );
    }

    return -IV_ERROR_UNREACHABLE;
}


// Uebertragen von Speicherdaten in Datei.
int ChannelBridge::writeMemToFile( uint8_t deviceNumber, int type, int address, int size, const std::string &filePath )
{
    // Remoteadresse einstellen
    setDeviceNumberReceiver( deviceNumber );

    // Aufgrund Protokollversion entscheiden
    // Protokollverwendung pruefen
    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2:
            return m_channel1_2.WriteMemToFile( type, address, size, filePath.c_str() );

        case 4:
            return m_channel4.writeMemToFile( deviceNumber, type, address, size, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Uebertragen von Speicherdaten aus einer Datei.
int ChannelBridge::writeFileToMem( uint8_t deviceNumber, const std::string &filePath )
{
    // Remoteadresse einstellen
    setDeviceNumberReceiver( deviceNumber );

    // Aufgrund Protokollversion entscheiden
    // Protokollverwendung pruefen
    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2:
            return m_channel1_2.WriteFileToMem( filePath );

        case 4:
            return m_channel4.writeFileToMem( deviceNumber, filePath );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Flash setzen/lesen (die ersten 16kByte Bootloader koennen nur gelesen werden)
int ChannelBridge::setFlash( uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    int err;
    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1: return m_channel1_2.WriteMem( address, count, buf, &err, 32 );
        case 2: return m_channel1_2.WriteMem( address, count, buf, &err );
        case 4: return m_channel4.writeToFlash( deviceNumber, address, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getFlash( uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    int err;
    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1: return m_channel1_2.ReadMem( address, count, buf, &err );
        case 2: return m_channel1_2.ReadMem( address, count, buf, &err );
        case 4: return m_channel4.readFromFlash( deviceNumber, address, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getFlashPage( uint8_t deviceNumber, uint32_t pageNum, uint32_t offs, uint8_t *buffer, uint32_t &count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readFlashPage( deviceNumber, pageNum, offs, buffer, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

// EEPROM setzen/lesen 
int ChannelBridge::setEEPROM( uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeToEeprom( deviceNumber, address, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getEEPROM( uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readFromEeprom( deviceNumber, address, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

// RAM setzen/lesen 
int ChannelBridge::setRAM( uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeToRam( deviceNumber, address, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getRAM( uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readFromRam( deviceNumber, address, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

// MMC setzen/lesen 
int ChannelBridge::setMMC( uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeToMmc( deviceNumber, address, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getMMC( uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readFromMmc( deviceNumber, address, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Systembios aufrufen/verlassen
int ChannelBridge::enterBios( uint8_t deviceNumber, uint8_t protocolVersion )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( protocolVersion )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.enterBiosMode( deviceNumber );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::exitBios( uint8_t deviceNumber )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.exitBiosMode( deviceNumber );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Interface setzen/lesen (RS232, TCPIP...)
int ChannelBridge::setInterface( uint8_t deviceNumber, uint8_t type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeMainInterfaceType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getInterface( uint8_t deviceNumber, uint8_t &type )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readMainInterfaceType( deviceNumber, type );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int ChannelBridge::setBaud( uint8_t deviceNumber, uint32_t br )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeBaud( deviceNumber, br );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getBaud( uint8_t deviceNumber, uint32_t &br )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readBaud( deviceNumber, br );
    }

    return -IV_ERROR_UNREACHABLE;
}

// DeviceID setzen/lesen
int ChannelBridge::setDeviceIdBios( uint8_t deviceNumber, uint16_t id )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeDeviceNumber( deviceNumber, id );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getDeviceIdBios( uint8_t deviceNumber, uint16_t &id )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readDeviceNumber( deviceNumber, id );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Bootloader aktivieren
int ChannelBridge::enterBootloader( uint8_t deviceNumber, uint8_t protocolVersion )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( protocolVersion )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.enterBootloaderMode( deviceNumber );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Monitor Mode vom XPort
int ChannelBridge::enterXPortMonitor( uint8_t deviceNumber, char *data, uint16_t &count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.enterXportMonitorMode( deviceNumber, data, count );
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_ENTER_MONITOR;
}

// Monitor Mode vom XPort
int ChannelBridge::enterXPortMonitorExt( uint8_t deviceNumber, char *data, uint16_t &count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.enterXportMonitorModeExt( deviceNumber, data, count );
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_ENTER_MONITOR;
}

int ChannelBridge::commandXPort( uint8_t deviceNumber, char *buf, uint16_t &count )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.xportMonitorModeCommand( deviceNumber, buf, count );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::exitXPortMonitor( uint8_t deviceNumber )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.exitXportMonitorMode( deviceNumber );
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_EXIT_MONITOR;
}

// Softwareinformation der Anwendungsfirmware abfragen
int ChannelBridge::getSoftwareInfo( uint8_t deviceNumber, uint8_t *swi, int &infoLength )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.ModuleTable::readSoftwareInfo( deviceNumber, swi, infoLength );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Oeffnen einer Tabelle
int ChannelBridge::tableOpen( uint8_t deviceNumber, const std::string &name, uint32_t options, int &handle, std::string &reserved )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.openTable( deviceNumber, name, options, handle, reserved );
    }

    return -IV_ERROR_UNREACHABLE;
}
// Schliessen der Tabelle
int ChannelBridge::tableClose( uint8_t deviceNumber, int handle )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.closeTable( deviceNumber, handle );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Setzen eines Filters auf die Tabelle
int ChannelBridge::tableSetFilter( uint8_t deviceNumber, int handle, const std::string &filter )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeTableFilter( deviceNumber, handle, filter );
    }

    return -IV_ERROR_UNREACHABLE;
}
// Lesen des Filters auf die Tabelle
int ChannelBridge::tableGetFilter( uint8_t deviceNumber, int handle, std::string &filter )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readTableFilter( deviceNumber, handle, filter );
    }

    return -IV_ERROR_UNREACHABLE;
}
// Loeschen des Filters der Tabelle
int ChannelBridge::tableRemoveFilter( uint8_t deviceNumber, int handle )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.removeTableFilter( deviceNumber, handle );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Ermitteln der Anzahl Datensaetze in der Tabelle
int ChannelBridge::tableGetRowCount( uint8_t deviceNumber, int handle, uint32_t &currentCount, uint32_t &unsortedCount, uint32_t &deletedCount )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readTableRowCount( deviceNumber, handle, currentCount, unsortedCount, deletedCount );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Ermitteln der Anzahl Datensaetze in der Tabelle
int ChannelBridge::tableGetCurrentRow( uint8_t deviceNumber, int handle, uint32_t &currentRow )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readCurrentTableRow( deviceNumber, handle, currentRow );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Ueberspringen von n Datensaetzen (offset kann negativ sein), ab top/aktuell/bottom
int ChannelBridge::tableSetCurrentRow( uint8_t deviceNumber, int handle, int offset, uint16_t direction )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeCurrentTableRow( deviceNumber, handle, offset, direction );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableSetCurrentRowData( uint8_t deviceNumber, int handle, uint16_t flags, char separator, const char *data, uint16_t dataSize )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeCurrentTableRowDataExt( deviceNumber, handle, flags, separator, data, dataSize );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableSetCurrentColumnData( uint8_t deviceNumber, int handle, uint16_t flags, const char *name, const char *data, uint16_t dataSize )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeCurrentTableColumnDataExt( deviceNumber, handle, flags, name, data, dataSize );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableSetAllRowsToColumnData( uint8_t deviceNumber, int handle, uint16_t flags, const char *name, const char *data, uint16_t dataSize )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeAllTableRowsColumnData( deviceNumber, handle, flags, name, data, dataSize );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abfragen eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableGetCurrentRowData( uint8_t deviceNumber, int handle, uint16_t flags, char separator, char *data, uint16_t &dataSize )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readCurrentTableRowDataExt( deviceNumber, handle, flags, separator, data, dataSize );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abfragen eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableGetCurrentColumnData( uint8_t deviceNumber, int handle, uint16_t flags, const char *name, char *data, uint16_t &dataSize )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readCurrentTableColumnDataExt( deviceNumber, handle, flags, name, data, dataSize );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableAppendRowData( uint8_t deviceNumber, int handle, uint16_t flags, char separator, const char *data, uint16_t dataSize )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.appendTableRowData( deviceNumber, handle, flags, separator, data, dataSize );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Loescht den aktuellen Datensatz
int ChannelBridge::tableDeleteCurrentRow( uint8_t deviceNumber, int handle, uint16_t flags )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.deleteCurrentTableRow( deviceNumber, handle, flags );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Loescht dee aktuellen Datensaetze, welche durch das gesetzte Filterkriterium verfuegbar sind.
int ChannelBridge::tableDeleteAvailableRows( uint8_t deviceNumber, int handle, uint16_t flags )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.deleteFiltratedTableRows( deviceNumber, handle, flags );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int ChannelBridge::getInfo( uint8_t deviceNumber, const std::string &name, const std::string &params, std::string &info )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readModuleInfo( deviceNumber, name, params, info );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Zutrittsbuchung durchfuehren
int ChannelBridge::entrance2Identification( uint8_t deviceNumber, const char *TM, const char *Id, const char *Pin, char *status, int &statusLength )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.accessControlIdentification( deviceNumber, TM, Id, Pin, status, statusLength );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::entrance2OnlineAction( uint8_t deviceNumber, const char *TM, uint32_t mask, uint8_t type, uint8_t duration )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.accessControlOnlineAction( deviceNumber, TM, mask, type, duration );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setNotifyOnDevice( uint8_t deviceNumber, int type, int state )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeNotify( deviceNumber, type, state );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz an Fingerprint-Tabelle anfuegen.
int ChannelBridge::fingerprintAppendRecord( uint8_t deviceNumber, uint8_t type, const uint8_t *fingertemplate, uint16_t size )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.appendFingerprintRecord( deviceNumber, type, fingertemplate, size );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle lesen.
int ChannelBridge::fingerprintGetRecord( uint8_t deviceNumber, uint8_t &type, uint32_t pid, uint8_t fid, uint8_t *fingertemplate, uint16_t &size )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readFingerprintRecord( deviceNumber, type, pid, fid, fingertemplate, size );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle loeschen.
int ChannelBridge::fingerprintDeleteRecord( uint8_t deviceNumber, uint32_t pid, uint8_t fid )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.deleteFingerprintRecord( deviceNumber, pid, fid );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Liste mit Fingerprintmoduldaten erzeugen.
int ChannelBridge::fingerprintList( uint8_t deviceNumber, int flags, uint32_t &recordCount )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.generateFingerprintIdList( deviceNumber, flags, recordCount );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Backup der Fingerprints erstellen
int ChannelBridge::fingerprintBackup( uint8_t deviceNumber, const std::string &filePath, int flags )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.fingerprintBackup( deviceNumber, filePath, flags );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Flags:
// Bit 1: gesetzt bedeutet: Enthaltene Fingertemplates beibehalten.
// Wiederherstellen der Fingerprints
int ChannelBridge::fingerprintRestore( uint8_t deviceNumber, const std::string &filePath, int flags )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.fingerprintRestore( deviceNumber, filePath, flags );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int ChannelBridge::setSystemFlags( uint8_t deviceNumber, uint32_t flags )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeSystemFlags( deviceNumber, flags );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSystemFlags( uint8_t deviceNumber, uint32_t &flags )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readSystemFlags( deviceNumber, flags );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionFirmware( uint8_t deviceNumber, int mask, int options )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.setOptionFirmware( deviceNumber, mask, options );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionFirmware( uint8_t deviceNumber, int &mask, int &options )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.getOptionFirmware( deviceNumber, mask, options );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Setzt den zu verwendenden Zeichensatztyp fuer sendMessage oder Sendinfotext
int ChannelBridge::setTypeFont( uint8_t deviceNumber, int iType, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            break;

        case 4: // PZEMasterIV mit Protokollversion 4
            // Befehlsausfuehrung starten.
            return resultToError( m_channel4.writeFontType( deviceNumberReceiver(), iType ), piError );

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Passwort pruefen
int ChannelBridge::passwordCheck( uint8_t deviceNumber, char *passwort )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeSetupPassword( deviceNumber, passwort );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int ChannelBridge::getPasswordKey( uint8_t deviceNumber, char *key )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readSetupPasswordKey( deviceNumber, key );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Tastendruck ueber Kommunikationsbefehl simulieren
int ChannelBridge::pressVirtualKey( uint8_t deviceNumber, uint8_t vk, uint8_t flags, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            break;

        case 4: // PZEMasterIV mit Protokollversion 4
            // Befehlsausfuehrung starten.
            return resultToError( m_channel4.writeVirtualKey( deviceNumberReceiver(), vk, flags ), piError );

        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getFlashStatus( uint8_t deviceNumber, int *piStatus, int *piError )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    uint32_t serial = 0;

    int iRes = 0;
    switch ( getProtocolVersion( deviceNumberReceiver() ) )
    {
        case 1:
        case 2:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            break;

        case 4:

            iRes = resultToError( m_channel4.readSerialNumber( deviceNumberReceiver(), serial ), piError );
            if ( iRes )
            {
                // Flags in Statusvariable uebernehmen
                *piStatus = (m_channel4.protocol().lastFlags() & 0x70) >> 4;
            }
            break;
        default:
            *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return iRes;
}

// Passwort fuer die verschluesselte Kommunikation
int ChannelBridge::setCommunicationPassword( const char *password, uint8_t size, uint32_t sessionTimeout, bool global )
{
    // Aufruf an Channel4 weiterleiten
    return m_channel4.protocol().setCommunicationPassword( password, size, sessionTimeout, global );
}

// Pruefen ob eine Globale Passwortvergabe moeglich ist
int ChannelBridge::isGlobalCommunicationPassword()
{
    // Aufruf an CChannel4 weiterleiten
    return m_channel4.protocol().aesCryptSession().isGlobalCommunicationPassword();
}

int ChannelBridge::recordVolume( uint8_t deviceNumber, int &number, int &memoryUsage, int &memorySize )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.readRecordsVolume( deviceNumber, number, memoryUsage, memorySize );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Funktionen für Pufferzugriff
int ChannelBridge::GetTableCount( uint8_t ucList )
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    uint8_t protocolVersion = (ucList) ? m_protocol_version_list_description : m_protocol_version_record_description;

    switch ( protocolVersion )
    {
        case 1:
        case 2:
            return m_channel1_2.GetTableCount( ucList );

        case 4:
            return m_channel4.GetTableCount( ucList );
    }

    return 0;
}

int ChannelBridge::GetTableRecord( uint8_t ucNum, std::string &szName, uint8_t *pucFieldCount, uint8_t *pucVariable, uint8_t ucList )
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    uint8_t protocolVersion = (ucList) ? m_protocol_version_list_description : m_protocol_version_record_description;

    // Bei Listen kann es nötig sein auf die Zutrittslistendaten zu verzweigen.
    if ( ucList && ucNum >= 100 )
    {
        // Auf Zutrtittskontrolllisten verzweigen
        if ( ucNum >= 100 && ucNum < 104 )
        {
// Version 1
            *pucVariable = 0;
            return m_channel1_2.GetTableRecordEntrance( ucNum, szName, pucFieldCount );
        }
        else if ( ucNum >= 200 && ucNum < 208 )
        {
// Version 2
            *pucVariable = 0;
            return m_channel4.GetTableRecordEntrance2( ucNum, szName, pucFieldCount );
        }
    }

    switch ( protocolVersion )
    {
        case 1:
        case 2:
            return m_channel1_2.GetTableRecord( ucNum, szName, pucFieldCount, pucVariable, ucList );

        case 4:
            return m_channel4.GetTableRecord( ucNum, szName, pucFieldCount, pucVariable, ucList );
    }

    return 0;
}

int ChannelBridge::GetTableField( uint8_t ucTableNum, uint8_t ucFieldNum, std::string &szName, uint8_t *pucType, uint8_t *pucLength, uint8_t ucList )
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    uint8_t protocolVersion = (ucList) ? m_protocol_version_list_description : m_protocol_version_record_description;

    // Bei Listen kann es nötig sein auf die Zutrittslistendaten zu verzweigen.
    if ( ucList && ucTableNum >= 100 )
    {
        // Auf Zutrittskontrolllisten verzweigen
        if ( ucTableNum >= 100 && ucTableNum < 104 )
        {
// Version 1
            return m_channel1_2.GetTableFieldEntrance( ucTableNum, ucFieldNum, szName, pucType, pucLength );
        }
        else if ( ucTableNum >= 200 && ucTableNum < 208 )
        {
// Version 2
            return m_channel4.GetTableFieldEntrance2( ucTableNum, ucFieldNum, szName, pucType, pucLength );
        }
    }

    switch ( protocolVersion )
    {
        case 1:
        case 2:
            return m_channel1_2.GetTableField( ucTableNum, ucFieldNum, szName, pucType, pucLength, ucList );

        case 4:
            return m_channel4.GetTableField( ucTableNum, ucFieldNum, szName, pucType, pucLength, ucList );
    }

    return 0;
}

void ChannelBridge::ClearListBuffer()
{
    // Importiert wird nur in Channel4, bei Load wird ggf. von Channel1_2 auf Channel4Buffer zugegriffen.
    m_channel4.ClearListBuffer();
}

int ChannelBridge::MakeListe( int ListNum, int ListCnt, int ListSize, const uint8_t *Buf )
{
    // Importiert wird nur in Channel4, bei Load wird ggf. von Channel1_2 auf Channel4Buffer zugegriffen.
    return m_channel4.MakeListe( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearEntranceListBuffer( int iListNum )
{
    m_channel1_2.ClearEntranceListBuffer( iListNum );
}

int ChannelBridge::MakeEntranceList( int ListNum, int ListCnt, int ListSize, const uint8_t *Buf )
{
    return m_channel1_2.MakeEntranceList( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearEntrance2ListBuffer( int iListNum )
{
    m_channel4.ClearEntrance2ListBuffer( iListNum );
}

int ChannelBridge::MakeEntrance2List( int ListNum, int ListCnt, int ListSize, const uint8_t *Buf )
{
    return m_channel4.MakeEntrance2List( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearTimeboyListBuffer( uint8_t id, uint8_t num )
{
    m_channel4.ClearTimeboyListBuffer( id, num );
}

int ChannelBridge::MakeTimeboyList( uint8_t id, uint8_t num, uint16_t recordCount, const uint8_t *buf, uint32_t size, int flags, int *error )
{
    return m_channel4.MakeTimeboyList( id, num, recordCount, buf, size, flags, error );
}

bool ChannelBridge::WriteListToBinFile( uint8_t listNum, const char *filePath )
{
    return m_channel4.WriteListToBinFile( listNum, filePath );
}

bool ChannelBridge::WriteListToTextFile( uint8_t listNum, const char *filePath )
{
    return m_channel4.WriteListToTextFile( listNum, filePath );
}

bool ChannelBridge::WriteListEntrance2ToBinFile( uint8_t listNum, const char *filePath )
{
    return m_channel4.WriteListEntrance2ToBinFile( listNum, filePath );
}

bool ChannelBridge::WriteListEntrance2ToTextFile( uint8_t listNum, const char *filePath )
{
    return m_channel4.WriteListEntrance2ToTextFile( listNum, filePath );
}

bool ChannelBridge::WriteListTimeboyToBinFile( uint8_t listId, uint8_t listNum, const char *filePath )
{
    return m_channel4.WriteListTimeboyToBinFile( listId, listNum, filePath );
}

std::string ChannelBridge::GetFileDeviceFileName()
{
    return m_channel4.m_fileDevice.m_fileName;
}

std::string ChannelBridge::GetFileDeviceError()
{
    return m_channel4.m_fileDevice.m_errorExtended;
}

std::string ChannelBridge::GetFileDeviceInfo()
{
    return m_channel4.m_hardwareInfoAsText;
}

FileDevice &ChannelBridge::GetFileDevice()
{
    return m_channel4.m_fileDevice;
}

SoftwarebuildVersion &ChannelBridge::minimalUpdateVersion()
{
    return m_channel4.m_minimalUpdateVersion;
}

std::list<std::string> &ChannelBridge::blacklistVersions()
{
    return m_channel4.m_blacklistVersion;
}

#if !defined( DATAFOX_STUDIO_IV ) && !defined( DF_API_PRODUCTION )
// Funktions-Stumps

// MODULE_HARDWARE

int ChannelBridge::setTypeDevice( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeBoard( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeTCPIP( uint8_t deviceNumber, uint8_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeModemMC( uint8_t deviceNumber, uint8_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeDisplay( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeKeypad( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeTransponder( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeMMC( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeRadio( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeGPS( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeDCF77( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeFingerprint( uint8_t deviceNumber, uint8_t type )
{
    (void)deviceNumber;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeExternalReader( uint8_t deviceNumber, uint8_t val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeExternalTransponderReaderType( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setComponentRS485_0( uint8_t deviceNumber, uint8_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentRS485_1( uint8_t deviceNumber, uint8_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentUSB( uint8_t deviceNumber, uint8_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentInAnalog( uint8_t deviceNumber, uint32_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentInDigital( uint8_t deviceNumber, uint32_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentOutDigital( uint8_t deviceNumber, uint32_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setOptionIOModule( uint8_t deviceNumber, uint8_t val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeIoModuleOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionTBAdaption( uint8_t deviceNumber, uint8_t val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeTimeboyAdaptionOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;

}

int ChannelBridge::setOptionComServer( uint8_t deviceNumber, uint8_t val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeComServerOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionTPWrite( uint8_t deviceNumber, uint8_t val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeTransponderWriteOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionHTTP( uint8_t deviceNumber, uint8_t val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeHttpOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionLanguage( uint8_t deviceNumber, uint8_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setOptionBiokeyUpdate( uint8_t deviceNumber, uint8_t val )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.writeFingerprintScannerUpdateOption( deviceNumber, val );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setPayableOptions( uint8_t deviceNumber, uint8_t mask, uint8_t options )
{
    (void)deviceNumber;
    (void)mask;
    (void)options;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setHWConfig( uint8_t deviceNumber, uint8_t val )
{
    (void)deviceNumber;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

// MODULE_SYSTEM

int ChannelBridge::setSerialNumber( uint8_t deviceNumber, uint32_t serial )
{
    (void)deviceNumber;
    (void)serial;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

#endif

// Zutrittsbuchung durchfuehren
int ChannelBridge::accessControlIdentification( uint8_t deviceNumber, int masterId, const char *TM, const char *Id, const char *Pin, char *status, int &statusLength )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.accessControlIdentificationExt( deviceNumber, masterId, TM, Id, Pin, status, statusLength );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::accessControlOnlineAction( uint8_t deviceNumber, int masterId, const char *TM, uint32_t mask, uint8_t type, uint8_t duration )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.accessControlOnlineActionExt( deviceNumber, masterId, TM, mask, type, duration );
    }

    return -IV_ERROR_UNREACHABLE;
}

// Zutrittskontroll-Befehl für Schließzylinder
int ChannelBridge::accessControlKnobCommand( uint8_t deviceNumber, uint8_t knobCommand, uint16_t masterId, uint8_t knobId, const char *params, uint8_t *data, int *sizeData )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.accessControlKnobCommand( deviceNumber, knobCommand, masterId, knobId, params, data, sizeData );
    }

    return -IV_ERROR_UNREACHABLE;
}


// Benutzerdefinierter Befehl
int ChannelBridge::customCommand( uint8_t deviceNumber, uint32_t idendification, uint32_t customCommand, const char *dataIn, int sizeIn, char *dataOut, int *sizeOut )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2: return -IV_ERROR_VERSION_FIRMWARE;
        case 4: return m_channel4.customCommand( deviceNumber, idendification, customCommand, dataIn, sizeIn, dataOut, sizeOut );
    }

    return -IV_ERROR_UNREACHABLE;
}


int ChannelBridge::startTransparencyMode( uint8_t deviceNumber, uint16_t id, uint16_t flags, uint32_t sizeReadBuffer, uint32_t sizeWriteBuffer, int timeout, int *errorNumber )
{
    // Busnummer uebernehmen
    setDeviceNumberReceiver( deviceNumber );
    int iRes = 0;

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2:
        {
            *errorNumber = IV_ERROR_VERSION_FIRMWARE;
            break;
        }
        case 4:
        {
            iRes = m_channel4.enterTransparencyMode( deviceNumber, id, flags, sizeReadBuffer, sizeWriteBuffer, timeout );
            if ( iRes != IV_ERROR_SUCCESS )
            {
                // Spezielle Funktionsfehler auswerten, ansonsten nach Standard auflösen.
                switch ( (iRes >> 8) & 0xFF )
                {
                    case 1:
                        *errorNumber = IV_ERROR_ALLOC_MEM;
                        iRes = 0;
                        break;

                    case 2:
                        *errorNumber = IV_ERROR_TIMEBOY_IV;
                        iRes = 0;
                        break;

                    default:
                        iRes = resultToError( iRes, errorNumber );
                }
            }
            else
            {
                iRes = 1;
            }
            break;
        }

        default:
            *errorNumber = IV_ERROR_UNREACHABLE;
            break;
    }

    return iRes;
}

int ChannelBridge::stopTransparencyMode( uint8_t deviceNumber, int *errorNumber )
{
    int iRes = 0;

    switch ( getProtocolVersion( deviceNumber ) )
    {
        case 1:
        case 2:
        {
            *errorNumber = IV_ERROR_VERSION_FIRMWARE;
            break;
        }
        case 4:
        {
            iRes = m_channel4.exitTransparencyMode( deviceNumber );
            if ( iRes != IV_ERROR_SUCCESS )
            {
                iRes = resultToError( iRes, errorNumber );
            }
            else
            {
                iRes = 1;
            }
            break;
        }

        default:
            *errorNumber = IV_ERROR_UNREACHABLE;
            break;
    }

    return iRes;
}

int ChannelBridge::isTransparencyMode()
{
    return m_channel4.isTransparencyMode();
}

int ChannelBridge::ping( uint8_t deviceNumber, int attempts, int *errorNumber )
{
    // Einen Versuch mehr.
    attempts++;

    for ( int i = 1; i < attempts; ++i )
    {
        if ( m_channel4.ping( deviceNumber, nullptr, nullptr ) == IV_ERROR_SUCCESS )
            return i;
    }

    *errorNumber = IV_ERROR_UNREACHABLE;
    return 0;
}

