/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/protocol_crypt.h"
#include <stdio.h>
#include <stdlib.h>
#include <boost/format.hpp>

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param connectionId Nummer der Kanalklasse fuer z.B. Trace.
* @param deviceAddress Busnummer der Kanalklasse.
*/
Datafox::Communication::ProtocolCrypt::ProtocolCrypt( int32_t connectionNumber, uint8_t deviceAddress )
    : ProtocolBase( connectionNumber, deviceAddress ) 
{
    // Umgebung zuruecksetzen
    reset();

    // Zufallsgenerator starten
    std::srand( static_cast<uint32_t>( std::time( nullptr ) ) );

    // Halten der vorhergehenden Sendedaten, um bei Wiederholdungen diese wiederzuverwenden.
    m_packetNumberPrevious = 0xFF;
    memset( m_packetDataPrevious, 0, sizeof(m_packetDataPrevious) );
    m_packetDataSizePrevious = 0;
}

// ##################################################################
/**
* @brief Destruktor.
*
*/
Datafox::Communication::ProtocolCrypt::~ProtocolCrypt() 
{
}

//#########################################################
//####################### HELPER ##########################
//#########################################################

// ##################################################################
/**
* @brief Initialisiert die grundlegend benoetigten Membervariablen.
*/
void Datafox::Communication::ProtocolCrypt::reset()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    ProtocolBase::reset();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geraetenummer.
*/
// Modulbefehl senden und Antwort zurueckgeben
//  deviceAddress -> Adresse (Geraetenummer) des Empfaengers
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
int32_t Datafox::Communication::ProtocolCrypt::call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int32_t contextValue )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p, %u, %i )\n", __FILE__, __LINE__, __func__, context, module, func, buffer, count, contextValue );
    int repeat = 0;

    // Im Crypt-Mode Userpassword muss auch ein solches vorliegen!
    // Bei Ping-Befehl pruefung unterlassen.
    if ( (context == IV_CONTEXT_MASTER)
        && !((module == IV_MODULE_PROTOCOL) && (func == IV_FUNCTION_PING))
        && aesCryptSession().m_cryptMode == IV_PACKET_CRYPT_MODE_PASSPHRASE && !aesCryptSession().isValidUserKey() )
    {
        return -IV_ERROR_NEED_USER_PASSWORD;
    }

label_again:
    // Muss eine Schluesselaushandlung aufgrund folgender Faelle durchgefuehrt werden:
    // Es ist eine Anfrage (ausser Schluesselaustausch selbst) die verschluesselte Kommunikation wird unterstuetzt und
    // soll eingesetzt werden und 
    // 1. Es wurde noch kein Session-Key ausgehandelt.
    // 2. Der aktuelle Session-Key ist aufgrund eines Timeouts ungueltig.
    if ( context == IV_CONTEXT_MASTER
        && !(module == IV_MODULE_PROTOCOL_CRYPT && func == IV_FUNCTION_EXCHANGE_PUBLIC_KEY) && aesCryptSession().m_cryptMode > 1 
        && (aesCryptSession().m_sessionTime == 0 || (aesCryptSession().m_sessionTimeout && time( nullptr ) > (aesCryptSession().m_sessionTime + aesCryptSession().m_sessionTimeout))) )
    {
        if ( aesCryptSession().m_sessionTime )
        {
            trace( "%s(%i):WARNING(1):  %s() [ session t(time): %i, t(timeout): %i, t(time-timeout): %i ]\n", __FILE__, __LINE__, __func__, static_cast<int>( time( nullptr ) - aesCryptSession().m_sessionTime ), aesCryptSession().m_sessionTimeout, static_cast<int>( time( NULL ) - (aesCryptSession().m_sessionTime + aesCryptSession().m_sessionTimeout) ) );
        }

        // Aushandlung bereits durchgefuehrt und laenger als 9 Sekunden abgelaufen?
        // Dann muessen die Crypt-Routinen neu aufgesetzt werden!
        if ( aesCryptSession().m_sessionTime > 0 && (time( nullptr ) - (aesCryptSession().m_sessionTime + aesCryptSession().m_sessionTimeout)) > 9 )
        {
            aesCryptSession().m_sessionTime = 0; // Session abgelaufen.
            aesCryptSession().initCryptMode();
            trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, aesCryptSession().m_cryptMode, (aesCryptSession().m_sessionTime == 0) ? "no" : "yes" );
        }

        // Buffer fuer Folgebefehl zwischenspeichern,
        // da das hinter buf liegende Bytearray durch Aufruf von exchangeSessionKey veraendert wird.
        uint16_t lCount = count;
        uint8_t data[IV_PACKET_LENGTH_DATA];
        memset( data, 0, sizeof( data ) );
        memcpy( data, buffer, count );

        // Aushandlung eines Session-Key durchfuehren
        int lRes = exchangeSessionKey( deviceNumberReceiver() );
        if ( lRes == 0x0D00 || lRes == -IV_ERROR_NEED_USER_PASSWORD )
            return -IV_ERROR_NEED_USER_PASSWORD;

        // Aushandlung erfolgreich durchgefuehrt, Parameterdaten wiederherstellen.
        count = lCount;
        memcpy( buffer, data, count );
    }

    int res = ProtocolBase::call( context, module, func, buffer, count, contextValue );
    // PROT_MODULE_ERROR_DECRYPT                (0x0D00) // Entschluesselung des Paketinhaltes bring Fehler.
    //                                                      Ablauf des Session-Key, Neuaushandlung ist erforderlich.
    if ( res == 0x0D00 )
    {
        //if ( module == IV_MODULE_PROTOCOL_CRYPT && func == IV_FUNCTION_EXCHANGE_PUBLIC_KEY )
        //  return res;

        aesCryptSession().m_sessionTime = 0; // Session-Key ungueltig.
        aesCryptSession().initCryptMode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %u, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, aesCryptSession().m_cryptMode, (aesCryptSession().m_sessionTime == 0) ? "no" : "yes" );
        if ( repeat == 0 )
        {
            repeat++;
            goto label_again;
        }
        else
        {
            // Das aktuell angegebene User-Passwort ist ungueltig.
            return -IV_ERROR_NEED_USER_PASSWORD;
        }
    }

    // PROT_MODULE_ERROR_CRYPTED                (0x0E00) // Mode passt nicht. Z. B. Geraet arbeitet verschluesselt und Anfrage war unverschluesselt.
    if ( res == 0x0E00 )
    {
        // Fehler bei Entschluesselung der Daten.
        aesCryptSession().m_cryptMode = 0;
        aesCryptSession().m_sessionTime = 0;
    }

    return res;
}

int32_t Datafox::Communication::ProtocolCrypt::writeProtocolCrypt()
{
    // Handelt es sich um die vorhergehende Paketnummer?
    bool resend = (m_packetNumberPrevious == m_packetSend[IV_PACKET_LENGTH_HEAD]);
    // Pakettyp uebernehmen
    //  unsigned char type = _packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3]];
    // Anzahl der Datenbyte
    uint16_t count = *(uint16_t *)&m_packetSend[4] - IV_PACKET_LENGTH_HEAD - m_packetSend[3] - sizeof( uint8_t ) - sizeof( uint16_t ) * 2;

    // Aktuell intern verwendeten Crypt-Mode eintragen.
    m_packetSend[6] |= aesCryptSession().m_cryptMode;

    // Wenn cryptMode > 1 (dann wird die Verschluesselung von der Gegenseite unterstuetzt),
    // dann verschluesseln und Verschluesselte Daten fuer ggf. wiederholtes
    // senden halten.
    if ( aesCryptSession().m_cryptMode > 1 )
    {
        uint16_t size;
        // Daten werden verschluesselt.
        m_packetSend[6] |= IV_PACKET_FLAG_CRYPTED;

        // Wiederholtes Senden?
        if ( resend )
        {
            // Zwischengespeicherten Dateninhalt wiederholt senden.
            memcpy( &m_packetSend[IV_PACKET_LENGTH_HEAD + m_packetSend[3] + sizeof( uint8_t )], m_packetDataPrevious, m_packetDataSizePrevious );
            size = m_packetDataSizePrevious;
        }
        else
        {
            size = aesCryptSession().sendEncrypt( &m_packetSend[IV_PACKET_LENGTH_HEAD + m_packetSend[3] + sizeof( uint8_t )], sizeof( uint16_t ) * 2 + count );
            // Aktuelle Daten fuer ggf. wiederholtes Senden halten.
            memcpy( m_packetDataPrevious, &m_packetSend[IV_PACKET_LENGTH_HEAD + m_packetSend[3] + sizeof( uint8_t )], size );
            m_packetDataSizePrevious = size;
        }

        // Anzahl Bytes die fuer die Verschluesselung hinzugefuegt wurde (16Byte Bloecke) berechnen.
        size -= sizeof( uint16_t ) * 2 + count;
        // Auf die Gesamtgroesse des Pakets draufaddieren.
        *(uint16_t *)&m_packetSend[4] += size;
        m_packetNumberPrevious = m_packetSend[IV_PACKET_LENGTH_HEAD];
    }

    return IV_ERROR_SUCCESS;
}

int32_t Datafox::Communication::ProtocolCrypt::writeStreamCrypt( std::string &buffer )
{
    // Wenn cryptMode > 1 (dann wird die Verschluesselung von der Gegenseite unterstuetzt),
    // dann verschluesseln und Verschluesselte Daten fuer ggf. wiederholtes
    // senden halten.
    if ( aesCryptSession().m_cryptMode > 1 )
    {
        uint32_t count_crypt_byte;
        uint32_t count_blocks;

        // Anzahl der Datenbyte
        uint32_t count = (uint32_t)buffer.size();

        // Anzahl 16 Byte Bloecke berechnen und Streamlaenge entsprechend anpassen.
        // Zusaetzlich werden die Daten um einen CRC und Anzahl Nutzdaten erweitert, daher (count + 5)!
        count_blocks = (count + 5) / 16;
        if ((count + 5) % 16)
            count_blocks++;

        count_crypt_byte = count_blocks * 16;

        // Puffer auf benoetigte Laenge anpassen.
        // Die Routine fuer Crypt fuellt den letzen 16 Byte Block des Puffer ggf. mit Zufallswerten auf.
        buffer.resize( count_crypt_byte, ' ' );

        // Datenblock verschluesseln.
        aesCryptSession().sendEncryptStream( (uint8_t *)&buffer[0], count );
    }

    return IV_ERROR_SUCCESS;
}

int32_t Datafox::Communication::ProtocolCrypt::readProtocolCrypt()
{
    // Pakettyp uebernehmen
    //  unsigned char type = _packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3]];
    // Anzahl der Datenbyte
    uint16_t count = *(uint16_t *)&m_packetReceive[4] - IV_PACKET_LENGTH_FRAME - m_packetReceive[3] - sizeof( uint8_t ) - sizeof( uint16_t ) * 2;

    // Wenn cryptMode > 1 (dann wird die Verschluesselung unterstuetzt), 
    // zusaetzlich die Daten verschluesselt vorliegen, dann entschluesseln.
    if ( (aesCryptSession().m_cryptMode > 1) && (m_lastFlags & IV_PACKET_FLAG_CRYPTED) )
    {
        uint16_t size = 2 * sizeof( uint16_t ) + count;
        if ( !aesCryptSession().receiveDecrypt( &m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3] + sizeof( uint8_t )], &size ) )
            return 0x0D00;

        *(uint16_t *)&m_packetReceive[4] -= count - (size - 2 * sizeof( uint16_t ));
        count = size - 2 * sizeof( uint16_t );
    }

    // Liegt noch keine Information ueber die unterstuetzte Verschluesselung der Gegenstelle vor,
    // muss diese ermittelt werden.
    if ( aesCryptSession().m_cryptMode != (m_lastFlags & 0x3) )
    {
        trace( "%s(%i):WARNING(1):  %s() [ cryptMode change %02X -> %02X ]\n", __FILE__, __LINE__, __func__, aesCryptSession().m_cryptMode, (m_lastFlags & 0x3) );

        // Die verschluesselte Kommunikation wird auf eine der folgenden Arten unterstuetzt.
        if ( (m_lastFlags & 0x3) == IV_PACKET_CRYPT_MODE_NO
            || (m_lastFlags & 0x3) == IV_PACKET_CRYPT_MODE_DEFAULT
            || (m_lastFlags & 0x3) == IV_PACKET_CRYPT_MODE_PASSPHRASE )
        {
            // Uebernahme des Verschluesselungsmode
            aesCryptSession().m_cryptMode = (m_lastFlags & 0x3);
        }
        else
        {
            // Die verschluesselte Kommunikation wird nicht unterstuetzt,
            // der Verschluesselungsmode ist somit 0.
            aesCryptSession().m_cryptMode = 0;
        }

        // Initialisieren des aktuell angenommenen Crypt-Mode.
        aesCryptSession().m_sessionTime = 0; // Eventuell aktueller Session-Key ist ungueltig.
        aesCryptSession().initCryptMode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, aesCryptSession().m_cryptMode, (aesCryptSession().m_sessionTime == 0) ? "no" : "yes" );
    }

    return IV_ERROR_SUCCESS;
}

int32_t Datafox::Communication::ProtocolCrypt::readStreamCrypt( std::string &buffer )
{
    // Wenn cryptMode > 1 (dann wird die Verschluesselung unterstuetzt), 
    // zusaetzlich die Daten verschluesselt vorliegen, dann entschluesseln.
    if ( aesCryptSession().m_cryptMode > 1 )
    {
        uint32_t size = (uint16_t)buffer.size();
        if ( !aesCryptSession().receiveDecryptStream( (uint8_t *)&buffer[0], &size ) )
            return 0x0D00;

        buffer.resize( size );
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Austausch eines Session-Key fuer die verschluesselte Kommunikation.
*
* @param deviceAddress Anzusprechende Geraetenummer.
*/
int32_t Datafox::Communication::ProtocolCrypt::exchangeSessionKey( uint8_t deviceNumber )
{
    trace( "%s(%i):INFO(3):  %s( %i )\n", __FILE__, __LINE__, __func__, deviceNumber );

    int result = -IV_ERROR_FAIL;
    uint16_t count = 0;
    uint8_t packet_data[IV_PACKET_LENGTH_DATA];

    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceNumber ) )
    {
        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver( deviceNumber );

        // Aufbau eines 8 Byte zufallsteils vom Session-Key
        uint8_t myKey[8];
        memset( myKey, 0, sizeof( myKey ) );
        uint8_t i;
        // Einen 8 Byte Langen Zufallswert erstellen.
        for ( i = 0; i < sizeof( myKey ); ++i )
            myKey[i] = rand() % 255;

        // Daten in Sendebuffer kopieren
        memcpy( packet_data, myKey, sizeof( myKey ) );
        count = sizeof( myKey );
        if ( aesCryptSession().m_cryptMode == IV_PACKET_CRYPT_MODE_PASSPHRASE )
            *(uint32_t *)&packet_data[count] = aesCryptSession().m_sessionTimeout;
        else
            *(uint32_t *)&packet_data[count] = 3600; // 1 Stunde default
        
        count += 4;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_PROTOCOL_CRYPT, IV_FUNCTION_EXCHANGE_PUBLIC_KEY, packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace( "%s(%i):INFO(2): %s() [ %04X ]\n", __FILE__, __LINE__, __func__, result & 0xFFFF );
            return result;
        }

        // Antwortparameter pruefen
        if ( count != 8 )
        {
            trace( "%s(%i):WARNING(1):  %s() [ Antwortparameter fehlt ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Aktuellen Aushandlungszeitpunkt halten
        aesCryptSession().m_sessionTime = time( NULL );

        // Gesamten Session-Key uebernehmen.
        memcpy( aesCryptSession().m_sessionKey, myKey, 8 );
        // Uebernehmen des Session-Key Teils vom Geraet.
        memcpy( &aesCryptSession().m_sessionKey[8], packet_data, 8 );
        aesCryptSession().m_sessionKeySize = 16;

        // Schluesselaushandlung erfolgreich.
        trace( "%s(%i):INFO(1): %s() [ Session-Key valid ]\n", __FILE__, __LINE__, __func__ );

        // Initialisieren der Verschluesselung (verwendet m_sessionTime, m_sessionKey)
        aesCryptSession().initCryptMode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, aesCryptSession().m_cryptMode, (aesCryptSession().m_sessionTime == 0) ? "no" : "yes" );
    }

    return result;
}

int32_t Datafox::Communication::ProtocolCrypt::setCommunicationPassword( const char* password, uint8_t size, uint32_t sessionTimeout, bool global )
{
    return aesCryptSession().setCommunicationPassword( password, size, sessionTimeout, global );
}
