/** 
* @file TimeboyStation.h
* @brief Schnittstelle fuer die Klasse TimeboyStation.
*/
//////////////////////////////////////////////////////////////////////
#pragma once
#ifndef DATAFOX_COMMUNICATION_TIMEBOY_STATION_V1_H
#define DATAFOX_COMMUNICATION_TIMEBOY_STATION_V1_H

// Basisklasse einbinden.
#include "datafox/communication/connection_base.h"

#include <time.h>
#include <thread>

//######################## TIMEBOY SLOT #######################################
// Defines Protokollbefehle
#define MB_READ		0x21		// bis zu 64 Byte lesen
#define MB_WRITE	0x22		// bis zu 64 Byte schreiben

// Defines fuer bestimmte Daten zu einem Slot
#define MB_SLOT_CHECK                  0x8000000C // Slot parameter fuer De/aktivierung der Timeboyerkennung
#define MB_SLOT_LED                    0x80001000 // Rote LED ein/aus
// Schreibe 1 Byte mit Wert 0 fuer: ein
// Schreibe 1 Byte mit Wert 255 fuer: aus
#define MB_SLOT_STATE                  0x8000E000 // Aktueller Steckzustand im Slot
#define MB_SLOT_VERSION                0x8000FFFF // Softwareversion der Slotfirmware

#define MB_SLOT_BUFFER_SIZE_SMALL      45  // Paketpuffergroesse bei FWVersion < 1.31
#define MB_SLOT_BUFFER_SIZE_LARGE      150 // Paketpuffergroesse bei FWVersion >= 1.31
#define MB_SLOT_BUFFER_DATA_SIZE_SMALL	(MB_SLOT_BUFFER_SIZE_SMALL - 13)
#define MB_SLOT_BUFFER_DATA_SIZE_LARGE	(MB_SLOT_BUFFER_SIZE_LARGE - 13)
//#############################################################################

// Versenden und Empfangen von Datenpaketen Tracen
//#define TRACE_DOCKING 1

namespace Datafox
{
    namespace Communication
    {
        // Datenstrucktur mit Daten zu einem Slot.
        class TimeboySlot  
        {
        public:
            TimeboySlot()
            {
                reset();
            }

        public:
            void reset()
            {
                // Member initialisieren
                m_version = 0;
                m_detectionState = 0;
                m_ledState = 0;
                m_ledStamp = 0;
                m_receiveStampSlot = 0;
                m_receiveStampTimeboy = 0;
            }
        private:
            uint16_t m_version;          // Ermittelte Firmwareversion
        public:
            void setVersion( uint16_t version )
            {
                m_version = version;
            }
            uint16_t version()
            {
                return m_version;
            }
        private:
            uint8_t m_detectionState;    // Ermittlung der Komperatorschaltung; 0 -> nicht ermittelt, 1 -> nicht verfuegbar, 2 -> verfuegbar
        public:
            void setDetectionState( uint8_t detectionState )
            {
                m_detectionState = detectionState;
            }
            uint8_t detectionState()
            {
                return m_detectionState;
            }
        private:
            uint8_t m_ledState;          // Zuletzt gesetzten LED-Zustand: 0x00 -> NA, 0x01 -> ON, 0xFF -> OFF
        public:
            void setLedState( uint8_t ledState )
            {
                m_ledState = ledState;
            }
            uint8_t ledState()
            {
                return m_ledState;
            }
        private:
            clock_t m_ledStamp;                // clock() des letzten Zustand setzen.
        public:
            void setLedStamp( clock_t ledStamp )
            {
                m_ledStamp = ledStamp;
            }
            clock_t ledStamp()
            {
                return m_ledStamp;
            }
        private:
            clock_t m_receiveStampSlot;        // clock() des zuletzt gueltig empfangenen Pakets.
        public:
            void setReceiveStampSlot( clock_t receiveStampSlot )
            {
                m_receiveStampSlot = receiveStampSlot;
            }
            clock_t receiveStampSlot()
            {
                return m_receiveStampSlot;
            }
        private:
            clock_t m_receiveStampTimeboy;     // clock() des zuletzt gueltig empfangene Antwort vom Timeboy
        public:
            void setReceiveStampTimeboy( clock_t receiveStampTimeboy )
            {
                m_receiveStampTimeboy = receiveStampTimeboy;
            }
            clock_t receiveStampTimeboy()
            {
                return m_receiveStampTimeboy;
            }

        };

// Automatische Jobs des Kommunikationsworkers
#define TSF_AUTO_LED         fb_1 // Ausschalten der Slot-LED durch Worker, nach vorgegebenem Timeout
#define TSF_AUTO_CLOSE       fb_2 // Schliessen der Schnittstelle, nach Abarbeitung aller Aufgaben.

        // Klasse fuer die Verwaltung von Slots
        class ProtocolDockingstationV1
        {
        public:
            enum
            {
                dv_write_trials = 3,
                dv_read_trials = 3
            };
        public:
            ProtocolDockingstationV1()
            {
                // Member initialisiseren.
                m_connection = nullptr;

                // Verbindungsthread
                m_threadState = ts_end;
                m_threadCommand = tc_none;

                // Anzahl der Suspendanforderungen halten
                m_suspend_counter = 0;

                // Flags fuer Steuerung der Idlezeit
                m_flags = 0;
            }
            virtual ~ProtocolDockingstationV1();

        public:
            // Setzt die Basisklasse fuer die Kommunikationsschicht
            void setConnection( ConnectionBase& connection )
            {
                m_connection = &connection;
            }
            ConnectionBase& connection() 
            { 
                if ( m_connection == nullptr )
                {
                    throw; 
                }

                return (ConnectionBase&)*m_connection; 
            }

            inline int32_t isSlot( uint8_t deviceNumber )
            {
                return (readSlotFirmwareVersion( deviceNumber ) > 0);
            }

            uint8_t bufferSize( uint8_t deviceNumber )
            {
                // Bereich pruefen
                if ( deviceNumber < 1 )
                {
                    return 0;
                }

                // Je nach enthaltener Version
                switch ( m_timeboySlot[deviceNumber].version() ) 
                {
                case 0xFFFF: // Pruefung bereits durchgefuehrt und war nicht erfolgreich.
                    break;

                case 0: // Ermittlung noch nicht durchgefuehrt
                    return MB_SLOT_BUFFER_SIZE_SMALL;

                default:
                    if ( m_timeboySlot[deviceNumber].version() < 0x0131 )
                    {
                        return MB_SLOT_BUFFER_SIZE_SMALL;
                    }
                    else if ( m_timeboySlot[deviceNumber].version() >= 0x0131 )
                    {
                        return MB_SLOT_BUFFER_SIZE_LARGE;
                    }
                }

                return 0;
            }
            uint8_t dataBufferSize( uint8_t deviceNumber )
            {
                // Bereich pruefen
                if ( deviceNumber < 1 )
                {
                    return 0;
                }

                switch ( m_timeboySlot[deviceNumber].version() ) 
                {
                case 0xFFFF: // Pruefung bereits durchgefuehrt und war nicht erfolgreich.
                    break;

                case 0: // Ermittlung noch nicht durchgefuehrt
                    return MB_SLOT_BUFFER_DATA_SIZE_SMALL;

                default:
                    if ( m_timeboySlot[deviceNumber].version() < 0x0131 )
                    {
                        return MB_SLOT_BUFFER_DATA_SIZE_SMALL;
                    }
                    else if ( m_timeboySlot[deviceNumber].version() >= 0x0131 )
                    {
                        return MB_SLOT_BUFFER_DATA_SIZE_LARGE;
                    }
                }

                return 0;
            }

            // Main - Funktion fuer Befehlsausfuehrung in Workerthread
        protected:
            void communicationThread();
            std::thread m_communicationThread;

        protected:
            // Aktueller interner Zustand des Threads halten
            volatile enum ThreadState 
            {
                ts_begin, // Bei Einstieg in Main-Routine
                ts_end,   // Bei Ausstieg aus Main-Routine
                ts_idle,  // Thread befindet sich im Leerlauf (laufende Pruefung auf anstehende Aufgaben)
                ts_execute_command, // Abarbeitung eines gesetzten Kommandos (threadCommand != none)
                ts_dummy
            } m_threadState;
        public:
            ThreadState threadState()
            {
                return m_threadState;
            }
        protected:
            // Auszufuehrendes Kommando
            volatile enum ThreadCommand
            {
                tc_none,  // Aktuell kein abzuarbeitender Befehl
                tc_exit,  // Beendigung des Threads erfordert
                tc_write, // Schreiben von Bytes erfordert (Daten in _commandParam)
                tc_read   // Lesen von Bytes erfordert (Daten in _commandParam)
            } m_threadCommand;
        public:
            ThreadCommand threadCommand()
            {
                return m_threadCommand;
            }
            int32_t setThreadCommand( ThreadCommand tc )
            {
                if ( m_threadState == ts_idle )
                {
                    m_threadCommand = tc;
                    return 1;
                }

                return 0;
            }
            volatile struct  
            {
                int32_t res;
                uint8_t deviceNumberReceiver;
                uint32_t baseAddress;
                int32_t offset;
                uint8_t* buffer;
                int32_t count;
                int32_t trials;
            } m_commandParameter;
            volatile int32_t m_suspend_counter;
        protected:
            int32_t startCommunicationThread();
            void stopCommunicationThread();

            // Gibt die FWVersion des Dockingslots an
            uint16_t readSlotFirmwareVersion( uint8_t deviceNumber );
        public:
            void resetSlot( uint8_t deviceNumber );
            void resetSlots();

        public: // Routinen fuer die Kommunikation mit Docking
            int32_t writeBytes( uint8_t deviceNumber, uint32_t baseAddress, int32_t offset, uint8_t *buffer, int32_t count, int32_t writeTrials = dv_write_trials );
            int32_t readBytes( uint8_t deviceNumber, uint32_t baseAddress, int32_t offset, uint8_t *buffer, int32_t count, int32_t readTrials = dv_read_trials );
        protected:
            int32_t threadWriteBytes( uint8_t deviceNumber, uint32_t baseAddress, int32_t offset, uint8_t *buffer, int32_t count, int32_t writeTrials = dv_write_trials );
            int32_t threadReadBytes( uint8_t deviceNumber, uint32_t baseAddress, int32_t offset, uint8_t *buffer, int32_t count, int32_t readTrials = dv_read_trials );

        public:
            int32_t readTimeboyModel( uint8_t deviceNumber, int32_t& model );
            int32_t setTimeboySlotState( uint8_t deviceNumber, int32_t state, int32_t commandResult = IV_ERROR_SUCCESS );

        protected:
            int32_t callDocking( uint8_t *sendBuffer, uint8_t sendCount, uint8_t *receiveBuffer, uint8_t *receiveCount );
            int32_t setDockingLed( uint8_t deviceNumber, int32_t state );

        protected: // Routinen fuer die Flagverwaltung
            volatile uint32_t m_flags;
        public:
            enum FlagBit
            {
                fb_1  = 0x00000001, fb_2  = 0x00000002, fb_3  = 0x00000004, fb_4  = 0x00000008, fb_5  = 0x00000010, fb_6  = 0x00000020, fb_7  = 0x00000040, fb_8  = 0x00000080,
                fb_9  = 0x00000100, fb_10 = 0x00000200, fb_11 = 0x00000400, fb_12 = 0x00000800, fb_13 = 0x00001000, fb_14 = 0x00002000, fb_15 = 0x00004000, fb_16 = 0x00008000,
                fb_17 = 0x00010000, fb_18 = 0x00020000, fb_19 = 0x00040000, fb_20 = 0x00080000, fb_21 = 0x00100000, fb_22 = 0x00200000, fb_23 = 0x00400000, fb_24 = 0x00800000,
                fb_25 = 0x01000000, fb_26 = 0x02000000, fb_27 = 0x04000000, fb_28 = 0x08000000, fb_29 = 0x10000000, fb_30 = 0x20000000, fb_31 = 0x40000000, fb_32 = 0x80000000
            };
            void addFlag( FlagBit flag )
            {
                m_flags |= flag;
            }
            void delFlag( FlagBit flag )
            {
                m_flags &= ~flag;
            }
            int32_t isFlag( FlagBit flag )
            {
                return ((m_flags & flag) != 0);
            }

        protected:
            int32_t m_autoLedTimeout;
        public:
            void setLedAutoControl( int32_t enable, int32_t timeout )
            {
                m_autoLedTimeout = timeout;
                if ( enable )
                {
                    addFlag( TSF_AUTO_LED );
                }
                else
                {
                    delFlag( TSF_AUTO_LED );
                }
            }

        protected:
            ConnectionBase* m_connection;
            TimeboySlot m_timeboySlot[255];

            // Aktuell bearbeitete DeviceID halten
        protected:
            uint8_t m_deviceNumberReceiver;
            inline void setDeviceNumberReceiver( uint8_t deviceNumber )
            {
                m_deviceNumberReceiver = deviceNumber;
            }
            inline uint8_t deviceNumberReceiver()
            {
                return m_deviceNumberReceiver;
            }

            // Helper zur Paketerstellung
        protected:
            int32_t binToUintMsbFirst( uint8_t *buffer )
            {
                return int32_t(buffer[0] << 8) | int32_t(buffer[1]);
            }
            uint32_t binToUlongMsbFirst( uint8_t *buffer )
            {
                uint32_t v;

                v  = buffer[0];
                v <<= 8;

                v |= buffer[1];
                v <<= 8;

                v |= buffer[2];
                v <<= 8;

                v |= buffer[3];

                return v;
            }

            uint16_t crc16( uint8_t *data, uint32_t size, uint16_t crc = 0xFFFF );
        };

    } // namespace Communication
} // namespace Datafox

#endif
